from typing import List

import torch
from rdkit import Chem

from .adj_mat_seer import AdjMatSeer
from .egnn import EGNNDynamics
from .equivariant_diffusion import (EquivariantDiffusion,
                                    PredefinedNoiseSchedule)
from .utils import (ATOM_DECODER, CONTEXT_NORMS, DIMENSION, MAX_N_NODES,
                    MIN_N_NODES, NUM_BOND_TYPES, align_mol_to_principal_frame,
                    apply_transform, extract_fragment,
                    prepare_adj_mat_seer_input, prepare_edm_input,
                    prepare_fragment, redefine_bonds, samples_to_rdkit_mol,
                    set_conformer_positions, standardize_mol)


class MLConformerGenerator(torch.nn.Module):
    """
    ML pipeline interface to generates novel molecules based on the 3D shape of a given reference molecule
    or an arbitrary context (principal components of MOI tensor).
    """

    def __init__(
        self,
        diffusion_steps: int = 100,
        device: torch.device | str = torch.device("cpu"),
        dimension: int = DIMENSION,
        num_bond_types: int = NUM_BOND_TYPES,
        min_n_nodes: int = MIN_N_NODES,
        max_n_nodes: int = MAX_N_NODES,
        context_norms: dict = CONTEXT_NORMS,
        atom_decoder: dict = ATOM_DECODER,
        edm_weights: str = "./edm_moi_chembl_15_39.pt",
        adj_mat_seer_weights: str = "./adj_mat_seer_chembl_15_39.pt",
    ):
        """
        Initialise the generator.

        :param diffusion_steps: Number of denoising steps - max 1000
        :param device: device to run the model on
        :param dimension: Maximal supported number of heavy atoms
        :param num_bond_types: Number of supported bond types
        :param min_n_nodes: Minimal value for number of heavy atoms in generated samples
        :param max_n_nodes: Maximal value for number of heavy atoms in generated samples
        :param context_norms: context normalisation parameters
        :param atom_decoder: decoder dict matching int atom encodings to string representations
        :param edm_weights: path to Equivariant Diffusion model state dict
        :param adj_mat_seer_weights: path to AdjMatSeer model state dict
        """
        super().__init__()

        self.device = device

        self.dimension = dimension
        self.atom_decoder = atom_decoder

        self.min_n_nodes = min_n_nodes
        self.max_n_nodes = max_n_nodes

        net_dynamics = EGNNDynamics(
            in_node_nf=9,
            context_node_nf=3,
            hidden_nf=420,
            device=device,
        )

        generative_model = EquivariantDiffusion(
            dynamics=net_dynamics,
            in_node_nf=8,
            timesteps=1000,
            noise_precision=1e-5,
        )

        adj_mat_seer = AdjMatSeer(
            dimension=dimension,
            n_hidden=2048,
            embedding_dim=64,
            num_embeddings=36,
            num_bond_types=num_bond_types,
            device=device,
        )

        gm_state_dict = torch.load(edm_weights, map_location=device)

        if "context_norms" in gm_state_dict:
            self.context_norms = {
                key: torch.tensor(value)
                for key, value in gm_state_dict["context_norms"].items()
            }
        else:
            self.context_norms = {
                key: torch.tensor(value) for key, value in context_norms.items()
            }

        generative_model.load_state_dict(gm_state_dict["state_dict"])

        ams_state_dict = torch.load(adj_mat_seer_weights, map_location=device)
        adj_mat_seer.load_state_dict(ams_state_dict["state_dict"])

        # Update denoising steps for the Equivarinat Diffusion
        generative_model.gamma = PredefinedNoiseSchedule(
            timesteps=diffusion_steps, precision=1e-5
        )

        generative_model.time_steps = torch.flip(
            torch.arange(0, diffusion_steps, device=device), dims=[0]
        )

        generative_model.T = diffusion_steps
        # ----------------------------

        generative_model.to(device)
        adj_mat_seer.to(device)

        generative_model.eval()
        adj_mat_seer.eval()

        self.generative_model = generative_model
        self.adj_mat_seer = adj_mat_seer

    @torch.inference_mode()
    def predict_bonds(self, edm_samples: list[Chem.Mol]) -> list[Chem.Mol]:
        """
        Predict bonds using the AdjMatSeer GCN model.
        :param edm_samples: List of RDKit molecule objects without bonds or with incorrect bonds.
        :return: List of RDKit molecule objects with predicted bonds.
        """

        (
            el_batch,
            dm_batch,
            b_adj_mat_batch,
            canonicalised_samples,
        ) = prepare_adj_mat_seer_input(
            mols=edm_samples,
            dimension=self.dimension,
            device=self.device,
        )

        adj_mat_batch = self.adj_mat_seer(
            elements=el_batch, dist_mat=dm_batch, adj_mat=b_adj_mat_batch
        )

        adj_mat_batch = adj_mat_batch.to("cpu")

        # Append generated bonds and standardise existing samples
        out_mols = []

        for i, adj_mat in enumerate(adj_mat_batch):
            mol = redefine_bonds(canonicalised_samples[i], adj_mat)
            out_mols.append(mol)

        return out_mols

    @torch.inference_mode()
    def edm_samples(
        self,
        reference_context: torch.Tensor,
        n_samples: int = 100,
        max_n_nodes: int = 32,
        min_n_nodes: int = 25,
        resample_steps: int = 0,
        fixed_fragment: Chem.Mol = None,
        blend_power: int = 3,
    ) -> List[Chem.Mol]:
        """
        Generates initial samples using the diffusion model
        :param reference_context: reference context - tensor of shape (3)
        :param n_samples: number of samples to be generated
        :param max_n_nodes: the maximal number of heavy atoms in the among requested molecules
        :param min_n_nodes: the minimal number of heavy atoms in the among requested molecules
        :param resample_steps: number of resampling steps applied for harmonisation of generation
        :param fixed_fragment: fragment to retain during generation, optional
        :param blend_power: power of polynomial blending of a fixed fragment during generation
        :return: a list of generated samples, without atom adjacency as RDkit Mol objects
        """

        # Make sure that number of atoms of generated samples is within allowed range
        if min_n_nodes < self.min_n_nodes:
            min_n_nodes = self.min_n_nodes

        if max_n_nodes > self.max_n_nodes:
            max_n_nodes = self.max_n_nodes

        node_mask, edge_mask, batch_context = prepare_edm_input(
            n_samples=n_samples,
            reference_context=reference_context,
            context_norms=self.context_norms,
            min_n_nodes=min_n_nodes,
            max_n_nodes=max_n_nodes,
            device=self.device,
        )

        if fixed_fragment is None:
            x, h = self.generative_model(
                node_mask,
                edge_mask,
                batch_context,
                resample_steps,
            )
        else:
            z_known, fixed_mask = prepare_fragment(
                n_samples=n_samples,
                fixed_fragment=fixed_fragment,
                max_n_nodes=max_n_nodes,
                min_n_nodes=min_n_nodes,
                device=self.device,
            )

            x, h = self.generative_model.inpaint(
                node_mask,
                edge_mask,
                batch_context,
                z_known,
                fixed_mask,
                resample_steps,
                blend_power,
            )

        mols = samples_to_rdkit_mol(
            positions=x, one_hot=h, node_mask=node_mask, atom_decoder=self.atom_decoder
        )

        return mols

    @torch.inference_mode()
    def generate_conformers(
        self,
        reference_conformer: Chem.Mol = None,
        n_samples: int = 10,
        variance: int = 2,
        reference_context: torch.Tensor = None,
        n_atoms: int = None,
        optimize_geometry: bool = True,
        resample_steps: int = 0,
        fixed_fragment: Chem.Mol | set = None,
        blend_power: int = 3,
    ) -> List[Chem.Mol]:
        """
        Main method to generate samples from either reference molecule or an arbitrary context.

        :param reference_conformer: A 3D conformer of a reference molecule as an RDKit Mol object
        :param n_samples: number of molecules to generate
        :param variance: int - variation in number of heavy atoms for generated molecules from reference
        :param reference_context: Arbitrary Reference context if applicable, instead of reference_conformer
        :param n_atoms: Reference number of atoms when generating using arbitrary context
        :param optimize_geometry: If true will apply constrained MMFF94 geometry optimisation to generated molecules
        :param resample_steps: number of resampling steps applied for harmonisation of generation
                               improves generation quality, while sacrificing speed
        :param fixed_fragment: Fragment to fix during generation as an RDKit Mol object or
                               a set of atom idxs of reference conformer
        :param blend_power: power of the polynomial blending schedule for generation with a fixed fragment
        :return: A list of valid standardised generated molecules as RDKit Mol objects
        """
        if reference_conformer:
            # Ensure the initial mol is stripped off Hs
            reference_conformer = Chem.RemoveAllHs(reference_conformer)
            ref_n_atoms = reference_conformer.GetNumAtoms()
            ref_context, shift, rotation, aligned_coord = align_mol_to_principal_frame(
                reference_conformer
            )

            if fixed_fragment:
                if isinstance(fixed_fragment, set):
                    aligned_ref_mol = set_conformer_positions(
                        reference_conformer, aligned_coord
                    )
                    fixed_fragment = extract_fragment(aligned_ref_mol, fixed_fragment)
                elif isinstance(fixed_fragment, Chem.Mol):
                    fixed_fragment = Chem.RemoveAllHs(fixed_fragment)
                    ff_conf = fixed_fragment.GetConformer()
                    ff_coord = torch.tensor(ff_conf.GetPositions(), dtype=torch.float32)
                    ff_coord_ref_aligned = apply_transform(ff_coord, shift, rotation)
                    fixed_fragment = set_conformer_positions(
                        fixed_fragment, ff_coord_ref_aligned
                    )

        elif reference_context is not None:
            if n_atoms:
                ref_n_atoms = n_atoms
            else:
                raise ValueError(
                    "Reference Number of Atoms should be provided, when generating samples using context."
                )

            ref_context = reference_context

            if isinstance(fixed_fragment, set):
                raise ValueError(
                    "'fixed_fragment' must be a Mol object when generating from a reference context."
                )

        else:
            raise ValueError(
                "Either a reference RDkit Mol object or context as torch.Tensor should be provided for generation."
            )

        edm_samples = self.edm_samples(
            reference_context=ref_context,
            n_samples=n_samples,
            min_n_nodes=ref_n_atoms - variance,
            max_n_nodes=ref_n_atoms + variance,
            resample_steps=resample_steps,
            fixed_fragment=fixed_fragment,
            blend_power=blend_power,
        )

        raw_mols = self.predict_bonds(edm_samples)

        # Append generated bonds and standardise existing samples
        optimised_conformers = []
        for f_mol in raw_mols:
            std_mol = standardize_mol(
                mol=f_mol, optimize_geometry=optimize_geometry, ifm_mode=False
            )
            if std_mol:
                optimised_conformers.append(std_mol)

        return optimised_conformers

    @torch.inference_mode()
    def forward(
        self,
        reference_conformer: Chem.Mol = None,
        n_samples: int = 10,
        variance: int = 2,
        reference_context: torch.Tensor = None,
        n_atoms: int = None,
        optimize_geometry: bool = True,
        resample_steps: int = 0,
        fixed_fragment: Chem.Mol | set = None,
        blend_power: int = 3,
    ) -> List[Chem.Mol]:
        return self.generate_conformers(
            reference_conformer=reference_conformer,
            n_samples=n_samples,
            variance=variance,
            reference_context=reference_context,
            n_atoms=n_atoms,
            optimize_geometry=optimize_geometry,
            resample_steps=resample_steps,
            fixed_fragment=fixed_fragment,
            blend_power=blend_power,
        )



from typing import Tuple

import torch
import torch.nn as nn


class GCL(nn.Module):
    """Graph Convolution layer based on aggregation"""

    def __init__(
        self,
        input_nf: int,
        output_nf: int,
        hidden_nf: int,
        normalization_factor: float = 100.0,
        edges_in_d: int = 0,
        nodes_att_dim: int = 0,
    ):
        super(GCL, self).__init__()
        input_edge = input_nf * 2
        self.normalization_factor = normalization_factor

        self.edge_mlp = nn.Sequential(
            nn.Linear(input_edge + edges_in_d, hidden_nf),
            nn.SiLU(),
            nn.Linear(hidden_nf, hidden_nf),
            nn.SiLU(),
        )

        self.node_mlp = nn.Sequential(
            nn.Linear(hidden_nf + input_nf + nodes_att_dim, hidden_nf),
            nn.SiLU(),
            nn.Linear(hidden_nf, output_nf),
        )

        self.att_mlp = nn.Sequential(nn.Linear(hidden_nf, 1), nn.Sigmoid())

    def edge_model(
        self,
        source: torch.Tensor,
        target: torch.Tensor,
        edge_attr: torch.Tensor,
        edge_mask: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        out = torch.cat([source, target, edge_attr], dim=1)
        mij = self.edge_mlp(out)

        att_val = self.att_mlp(mij)
        out = mij * att_val

        out = out * edge_mask
        return out, mij

    def node_model(
        self, x: torch.Tensor, edge_index: torch.Tensor, edge_attr: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        row = edge_index[0]

        agg = unsorted_segment_sum(
            data=edge_attr,
            segment_ids=row,
            num_segments=x.size(0),
            normalization_factor=self.normalization_factor,
        )

        agg = torch.cat([x, agg], dim=1)
        out = x + self.node_mlp(agg)
        return out, agg

    def forward(
        self,
        h: torch.Tensor,
        edge_index: torch.Tensor,
        edge_attr: torch.Tensor,
        node_mask: torch.Tensor,
        edge_mask: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        row = edge_index[0]
        col = edge_index[1]

        edge_feat, mij = self.edge_model(h[row], h[col], edge_attr, edge_mask)
        h, agg = self.node_model(h, edge_index, edge_feat)
        h = h * node_mask

        return h, mij


class EquivariantUpdate(nn.Module):
    def __init__(
        self,
        hidden_nf: int,
        normalization_factor: float = 100.0,
        edges_in_d: int = 1,
        coords_range: float = 10.0,
    ):
        super(EquivariantUpdate, self).__init__()

        self.coords_range = coords_range
        input_edge = hidden_nf * 2 + edges_in_d
        layer = nn.Linear(hidden_nf, 1, bias=False)
        nn.init.xavier_uniform_(layer.weight, gain=0.001)
        self.coord_mlp = nn.Sequential(
            nn.Linear(input_edge, hidden_nf),
            nn.SiLU(),
            nn.Linear(hidden_nf, hidden_nf),
            nn.SiLU(),
            layer,
        )
        self.normalization_factor = normalization_factor

    def coord_model(
        self,
        h: torch.Tensor,
        coord: torch.Tensor,
        edge_index: torch.Tensor,
        coord_diff: torch.Tensor,
        edge_attr: torch.Tensor,
        edge_mask: torch.Tensor,
    ) -> torch.Tensor:
        row = edge_index[0]
        col = edge_index[1]
        input_tensor = torch.cat([h[row], h[col], edge_attr], dim=1)

        trans = coord_diff * self.coord_mlp(input_tensor)

        if edge_mask is not None:
            trans = trans * edge_mask
        agg = unsorted_segment_sum(
            data=trans,
            segment_ids=row,
            num_segments=coord.size(0),
            normalization_factor=self.normalization_factor,
        )
        coord = coord + agg
        return coord

    def forward(
        self,
        h: torch.Tensor,
        coord: torch.Tensor,
        edge_index: torch.Tensor,
        coord_diff: torch.Tensor,
        edge_attr: torch.Tensor,
        node_mask: torch.Tensor,
        edge_mask: torch.Tensor,
    ) -> torch.Tensor:
        coord = self.coord_model(h, coord, edge_index, coord_diff, edge_attr, edge_mask)
        coord = coord * node_mask
        return coord


class EquivariantBlock(nn.Module):
    def __init__(
        self,
        hidden_nf: int = 2,
        edge_feat_nf: int = 2,
        coords_range: float = 15.0,
        normalization_factor: float = 100.0,
    ):
        super(EquivariantBlock, self).__init__()
        self.hidden_nf = hidden_nf
        self.coords_range_layer = coords_range
        self.normalization_factor = normalization_factor

        self.gcl_0 = GCL(
            input_nf=hidden_nf,
            output_nf=hidden_nf,
            hidden_nf=hidden_nf,
            edges_in_d=edge_feat_nf,
            normalization_factor=normalization_factor,
        )

        self.gcl_1 = GCL(
            input_nf=hidden_nf,
            output_nf=hidden_nf,
            hidden_nf=hidden_nf,
            edges_in_d=edge_feat_nf,
            normalization_factor=normalization_factor,
        )

        self.gcl_equiv = EquivariantUpdate(
            hidden_nf=hidden_nf,
            edges_in_d=edge_feat_nf,
            coords_range=self.coords_range_layer,
            normalization_factor=normalization_factor,
        )

    def forward(
        self,
        h: torch.Tensor,
        x: torch.Tensor,
        edge_index: torch.Tensor,
        node_mask: torch.Tensor,
        edge_mask: torch.Tensor,
        edge_attr: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        distances, coord_diff = coord2diff(x, edge_index)

        edge_attr = torch.cat([distances, edge_attr], dim=1)

        h, _ = self.gcl_0(
            h=h,
            edge_index=edge_index,
            edge_attr=edge_attr,
            node_mask=node_mask,
            edge_mask=edge_mask,
        )

        h, _ = self.gcl_1(
            h=h,
            edge_index=edge_index,
            edge_attr=edge_attr,
            node_mask=node_mask,
            edge_mask=edge_mask,
        )

        x = self.gcl_equiv(
            h, x, edge_index, coord_diff, edge_attr, node_mask, edge_mask
        )

        h = h * node_mask
        return h, x


class EGNN(nn.Module):
    def __init__(
        self,
        in_node_nf: int,
        hidden_nf: int,
        coords_range: float = 15.0,
        normalization_factor: float = 100.0,
    ):
        super(EGNN, self).__init__()
        self.hidden_nf = hidden_nf
        self.normalization_factor = normalization_factor

        edge_feat_nf = 2

        self.embedding = nn.Linear(in_node_nf, self.hidden_nf)
        self.embedding_out = nn.Linear(self.hidden_nf, in_node_nf)

        self.e_block_0 = EquivariantBlock(
            hidden_nf=hidden_nf,
            edge_feat_nf=edge_feat_nf,
            coords_range=coords_range,
            normalization_factor=self.normalization_factor,
        )

        self.e_block_1 = EquivariantBlock(
            hidden_nf=hidden_nf,
            edge_feat_nf=edge_feat_nf,
            coords_range=coords_range,
            normalization_factor=self.normalization_factor,
        )

        self.e_block_2 = EquivariantBlock(
            hidden_nf=hidden_nf,
            edge_feat_nf=edge_feat_nf,
            coords_range=coords_range,
            normalization_factor=self.normalization_factor,
        )

        self.e_block_3 = EquivariantBlock(
            hidden_nf=hidden_nf,
            edge_feat_nf=edge_feat_nf,
            coords_range=coords_range,
            normalization_factor=self.normalization_factor,
        )

        self.e_block_4 = EquivariantBlock(
            hidden_nf=hidden_nf,
            edge_feat_nf=edge_feat_nf,
            coords_range=coords_range,
            normalization_factor=self.normalization_factor,
        )

        self.e_block_5 = EquivariantBlock(
            hidden_nf=hidden_nf,
            edge_feat_nf=edge_feat_nf,
            coords_range=coords_range,
            normalization_factor=self.normalization_factor,
        )

        self.e_block_6 = EquivariantBlock(
            hidden_nf=hidden_nf,
            edge_feat_nf=edge_feat_nf,
            coords_range=coords_range,
            normalization_factor=self.normalization_factor,
        )

        self.e_block_7 = EquivariantBlock(
            hidden_nf=hidden_nf,
            edge_feat_nf=edge_feat_nf,
            coords_range=coords_range,
            normalization_factor=self.normalization_factor,
        )

        self.e_block_8 = EquivariantBlock(
            hidden_nf=hidden_nf,
            edge_feat_nf=edge_feat_nf,
            coords_range=coords_range,
            normalization_factor=self.normalization_factor,
        )

    def forward(
        self,
        h: torch.Tensor,
        x: torch.Tensor,
        edge_index: torch.Tensor,
        node_mask: torch.Tensor,
        edge_mask: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        distances, _ = coord2diff(x, edge_index)

        h = self.embedding(h)

        h, x = self.e_block_0(
            h=h,
            x=x,
            edge_index=edge_index,
            node_mask=node_mask,
            edge_mask=edge_mask,
            edge_attr=distances,
        )

        h, x = self.e_block_1(
            h=h,
            x=x,
            edge_index=edge_index,
            node_mask=node_mask,
            edge_mask=edge_mask,
            edge_attr=distances,
        )

        h, x = self.e_block_2(
            h=h,
            x=x,
            edge_index=edge_index,
            node_mask=node_mask,
            edge_mask=edge_mask,
            edge_attr=distances,
        )

        h, x = self.e_block_3(
            h=h,
            x=x,
            edge_index=edge_index,
            node_mask=node_mask,
            edge_mask=edge_mask,
            edge_attr=distances,
        )

        h, x = self.e_block_4(
            h=h,
            x=x,
            edge_index=edge_index,
            node_mask=node_mask,
            edge_mask=edge_mask,
            edge_attr=distances,
        )

        h, x = self.e_block_5(
            h=h,
            x=x,
            edge_index=edge_index,
            node_mask=node_mask,
            edge_mask=edge_mask,
            edge_attr=distances,
        )

        h, x = self.e_block_6(
            h=h,
            x=x,
            edge_index=edge_index,
            node_mask=node_mask,
            edge_mask=edge_mask,
            edge_attr=distances,
        )

        h, x = self.e_block_7(
            h=h,
            x=x,
            edge_index=edge_index,
            node_mask=node_mask,
            edge_mask=edge_mask,
            edge_attr=distances,
        )

        h, x = self.e_block_8(
            h=h,
            x=x,
            edge_index=edge_index,
            node_mask=node_mask,
            edge_mask=edge_mask,
            edge_attr=distances,
        )

        h = self.embedding_out(h)
        h = h * node_mask

        return h, x


def coord2diff(
    x: torch.Tensor, edge_index: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
    row = edge_index[0]
    col = edge_index[1]

    coord_diff = x[row] - x[col]
    radial = torch.sum(coord_diff**2, 1).unsqueeze(1)
    norm = torch.sqrt(radial + 1e-8)
    coord_diff = coord_diff / norm

    return radial, coord_diff


def unsorted_segment_sum(
    data: torch.Tensor,
    segment_ids: torch.Tensor,
    num_segments: int,
    normalization_factor: float,
) -> torch.Tensor:
    """
    Custom PyTorch op to replicate TensorFlow's `unsorted_segment_sum`.
    Normalization: 'sum'.
    """

    result = torch.zeros(
        (num_segments, data.size(1)), dtype=data.dtype, device=data.device
    )
    segment_ids = segment_ids.unsqueeze(-1).expand_as(data)

    result.scatter_add_(0, segment_ids, data)
    result = result / normalization_factor

    return result


def remove_mean_with_mask(x: torch.Tensor, node_mask: torch.Tensor) -> torch.Tensor:
    n = torch.sum(node_mask, 1, keepdim=True)

    mean = torch.sum(x, dim=1, keepdim=True) / n
    x = x - mean * node_mask
    return x


class EGNNDynamics(nn.Module):
    def __init__(
        self,
        in_node_nf: int,
        context_node_nf: int,  # -> Calculated from context properties
        n_dims: int = 3,
        hidden_nf: int = 420,  # -> 420 our default
        device: torch.device = torch.device("cpu"),
        normalization_factor: float = 100.0,
    ):
        super().__init__()

        # Should have 9 Equivarinat blocks n_layers = 9
        self.egnn = EGNN(
            in_node_nf=in_node_nf + context_node_nf,
            hidden_nf=hidden_nf,
            normalization_factor=normalization_factor,
        )
        self.in_node_nf = in_node_nf

        self.context_node_nf = context_node_nf
        self.device = device
        self.n_dims = n_dims

    def forward(self, t, xh, node_mask, edge_mask, context):
        bs, n_nodes, _ = xh.size()

        edges = self.get_adj_matrix(n_nodes, bs, self.device)

        node_mask = node_mask.view(bs * n_nodes, 1)
        edge_mask = edge_mask.view(bs * n_nodes * n_nodes, 1)
        xh = xh.view(bs * n_nodes, -1).clone() * node_mask
        x = xh[:, 0 : self.n_dims].clone()

        h = xh[:, self.n_dims :].clone()

        h_time = t.view(bs, 1).repeat(1, n_nodes)
        h_time = h_time.view(bs * n_nodes, 1)

        h = torch.cat([h, h_time], dim=1)

        # Context is added for conditional generation

        context = context.view(bs * n_nodes, self.context_node_nf)

        h = torch.cat([h, context], dim=1)

        h_final, x_final = self.egnn(
            h=h, x=x, edge_index=edges, node_mask=node_mask, edge_mask=edge_mask
        )

        vel = (x_final - x) * node_mask

        h_final = h_final[:, : -self.context_node_nf]

        h_final = h_final[:, :-1]

        vel = vel.view(bs, n_nodes, -1)

        vel = remove_mean_with_mask(vel, node_mask.view(bs, n_nodes, 1))

        h_final = h_final.view(bs, n_nodes, -1)

        return torch.cat([vel, h_final], dim=2)

    @staticmethod
    def get_adj_matrix(
        n_nodes: int, batch_size: int, device: torch.device
    ) -> torch.Tensor:
        # Generate batch offsets
        batch_offsets = torch.arange(batch_size, device=device).unsqueeze(1) * n_nodes

        # Generate row and column indices for a single batch
        row_indices = (
            torch.arange(n_nodes, device=device).repeat(n_nodes, 1).T.flatten()
        )
        col_indices = torch.arange(n_nodes, device=device).repeat(n_nodes)

        # Expand to all batches
        rows = (row_indices.unsqueeze(0) + batch_offsets).flatten()
        cols = (col_indices.unsqueeze(0) + batch_offsets).flatten()

        # Store the edges as LongTensor
        edges = torch.stack(
            [
                rows.long(),
                cols.long(),
            ],
            dim=0,
        ).to(device)

        return edges



"""
Shared constants and pure-RDKit functions used by both PyTorch and ONNX code paths.
"""

from rdkit import Chem
from rdkit.Chem import rdDetermineBonds
from rdkit.Geometry import Point3D

from .config import PERMITTED_ELEMENTS

elements_decoder = {x: i for i, x in enumerate(sorted(PERMITTED_ELEMENTS))}

# allowable node and edge features
allowable_features = {
    "possible_atomic_num_list": list(range(1, 35)),
    "possible_implicit_valence_list": [0, 1, 2, 3, 4, 5, 6],
    "possible_degree_list": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    "possible_bonds": [
        Chem.rdchem.BondType.SINGLE,
        Chem.rdchem.BondType.DOUBLE,
        Chem.rdchem.BondType.TRIPLE,
        Chem.rdchem.BondType.AROMATIC,
    ],
}

elements_dict = {
    1: "H",
    6: "C",
    7: "N",
    8: "O",
    9: "F",
    15: "P",
    16: "S",
    17: "Cl",
    35: "Br",
}

bond_type_dict = {
    1: Chem.rdchem.BondType.SINGLE,
    2: Chem.rdchem.BondType.DOUBLE,
    3: Chem.rdchem.BondType.TRIPLE,
    4: Chem.rdchem.BondType.AROMATIC,
}


def canonicalise(mol: Chem.Mol) -> Chem.Mol:
    """
    Bring order of atoms in the molecule to canonical based on generic one-order connectivity
    :param mol: Mol object with unordered atoms
    :return: Mol object with canonicalised order of atoms
    """
    # Guess simple 1-order connectivity and re-order the molecule
    rdDetermineBonds.DetermineConnectivity(mol)
    _ = Chem.MolToSmiles(mol)
    order_str = mol.GetProp("_smilesAtomOutputOrder")

    order_str = order_str.replace("[", "").replace("]", "")
    order = [int(x) for x in order_str.split(",") if x != ""]

    mol_ordered = Chem.RenumberAtoms(mol, order)

    return mol_ordered


def apply_transform(coord, shift, rotation):
    """
    Apply Translation -> Rotation transform to a set of coordinates
    :param coord: Coordinates
    :param shift: Shift (Translation) matrix
    :param rotation: Rotation matrix
    :returns: Transformed coordinates
    """
    coord_shifted = coord + shift
    coord_transformed = coord_shifted @ rotation
    return coord_transformed


def set_conformer_positions(mol, coord):
    conf = mol.GetConformer()
    for i, point in enumerate(coord):
        x, y, z = point.tolist()
        conf.SetAtomPosition(i, Point3D(x, y, z))

    return mol



from typing import Iterable

from rdkit import Chem

MIN_SIZE_DEFAULT = 6
MAX_SIZE_DEFAULT = 20


def find_cuttable_bonds(mol):
    """Single, non-ring bonds only."""
    cuttable = []
    for bond in mol.GetBonds():
        if bond.IsInRing():
            continue
        if bond.GetBondType() != Chem.BondType.SINGLE:
            continue
        cuttable.append(bond.GetIdx())
    return cuttable


def _build_fragment_adjacency(mol, fragments):
    atom_to_frag = {}
    for fi, atom_set in enumerate(fragments):
        for idx in atom_set:
            atom_to_frag[idx] = fi

    adj = {i: set() for i in range(len(fragments))}
    for bond in mol.GetBonds():
        a1 = bond.GetBeginAtomIdx()
        a2 = bond.GetEndAtomIdx()
        if a1 not in atom_to_frag or a2 not in atom_to_frag:
            continue
        f1 = atom_to_frag[a1]
        f2 = atom_to_frag[a2]
        if f1 != f2:
            adj[f1].add(f2)
            adj[f2].add(f1)
    return adj


def _components_after_cut(mol, atom_set, cut_bond_idx):
    local_adj = {a: set() for a in atom_set}
    for b in mol.GetBonds():
        a1 = b.GetBeginAtomIdx()
        a2 = b.GetEndAtomIdx()
        if a1 not in atom_set or a2 not in atom_set:
            continue
        if b.GetIdx() == cut_bond_idx:
            continue
        local_adj[a1].add(a2)
        local_adj[a2].add(a1)

    seen = set()
    comps = []
    for a in atom_set:
        if a in seen:
            continue
        stack = [a]
        comp = set()
        while stack:
            x = stack.pop()
            if x in seen:
                continue
            seen.add(x)
            comp.add(x)
            stack.extend(local_adj[x] - seen)
        comps.append(comp)

    return comps


def _try_merge_small_fragment(mol, fragments, min_size, max_size):
    sizes = [len(s) for s in fragments]
    adj = _build_fragment_adjacency(mol, fragments)

    small_idx = None
    for i, sz in enumerate(sizes):
        if sz < min_size:
            small_idx = i
            break
    if small_idx is None:
        return False, fragments

    neighbors = list(adj[small_idx])
    if not neighbors:
        return False, fragments  # isolated tiny bit

    target = 0.5 * (min_size + max_size)
    best_j = None
    best_score = 1e9

    for j in neighbors:
        merged_size = sizes[small_idx] + sizes[j]
        penalty = 0.0
        if merged_size < min_size:
            penalty += (min_size - merged_size) * 100.0
        if merged_size > max_size:
            penalty += (merged_size - max_size) * 10.0
        penalty += abs(merged_size - target)
        if penalty < best_score:
            best_score = penalty
            best_j = j

    if best_j is None:
        return False, fragments

    i, j = small_idx, best_j
    if j < i:
        i, j = j, i

    merged = fragments[i] | fragments[j]

    new_frags = []
    for k, fset in enumerate(fragments):
        if k == i:
            new_frags.append(merged)
        elif k == j:
            continue
        else:
            new_frags.append(fset)

    return True, new_frags


def _try_split_large_fragment(mol, fragments, cuttable_bonds, min_size, max_size):
    sizes = [len(s) for s in fragments]
    target = 0.5 * (min_size + max_size)

    large_idx = None
    for i, sz in enumerate(sizes):
        if sz > max_size:
            large_idx = i
            break
    if large_idx is None:
        return False, fragments

    atom_set = fragments[large_idx]
    candidates = []

    for bidx in cuttable_bonds:
        bond = mol.GetBondWithIdx(bidx)
        a1 = bond.GetBeginAtomIdx()
        a2 = bond.GetEndAtomIdx()
        if a1 not in atom_set or a2 not in atom_set:
            continue

        comps = _components_after_cut(mol, atom_set, bidx)
        if len(comps) != 2:
            continue

        s1 = len(comps[0])
        s2 = len(comps[1])

        penalty = 0.0
        for s in (s1, s2):
            if s < min_size:
                penalty += (min_size - s) * 100.0
            if s > max_size:
                penalty += (s - max_size) * 10.0
            penalty += abs(s - target)

        candidates.append((penalty, comps))

    if not candidates:
        return False, fragments

    candidates.sort(key=lambda x: x[0])
    comp1, comp2 = candidates[0][1]

    new_frags = []
    for i, fset in enumerate(fragments):
        if i == large_idx:
            new_frags.append(comp1)
            new_frags.append(comp2)
        else:
            new_frags.append(fset)

    return True, new_frags


def _submol_from_atom_set(mol, atom_set):
    amap = {}
    emol = Chem.RWMol()
    for old_idx in sorted(atom_set):
        atom = mol.GetAtomWithIdx(old_idx)
        new_atom = Chem.Atom(atom.GetAtomicNum())
        new_idx = emol.AddAtom(new_atom)
        amap[old_idx] = new_idx

    for bond in mol.GetBonds():
        a1 = bond.GetBeginAtomIdx()
        a2 = bond.GetEndAtomIdx()
        if a1 in atom_set and a2 in atom_set:
            emol.AddBond(amap[a1], amap[a2], bond.GetBondType())

    frag = emol.GetMol()
    Chem.SanitizeMol(frag)
    return frag


def split_molecule_size_constrained(
    mol,
    min_size=MIN_SIZE_DEFAULT,
    max_size=MAX_SIZE_DEFAULT,
    max_iter=200,
    verbose=False,
):
    heavy_atoms = [i for i, a in enumerate(mol.GetAtoms()) if a.GetAtomicNum() > 1]
    if len(heavy_atoms) <= min_size:
        if verbose:
            print("Molecule has <= min_size heavy atoms; single fragment.")
        return [set(heavy_atoms)], [_submol_from_atom_set(mol, set(heavy_atoms))]

    fragments = [set(heavy_atoms)]
    cuttable_bonds = find_cuttable_bonds(mol)
    if verbose:
        print("Cuttable bond indices:", cuttable_bonds)

    for it in range(max_iter):
        sizes = [len(s) for s in fragments]
        if verbose:
            print(f"Iter {it}: sizes = {sizes}")

        if all(min_size <= s <= max_size for s in sizes):
            if verbose:
                print("All fragments within size window; stopping.")
            break

        # changed = False

        if any(s > max_size for s in sizes):
            changed, fragments = _try_split_large_fragment(
                mol, fragments, cuttable_bonds, min_size, max_size
            )
            if changed:
                continue

        if any(s < min_size for s in sizes):
            changed, fragments = _try_merge_small_fragment(
                mol, fragments, min_size, max_size
            )
            if changed:
                continue

        if verbose:
            print("No further merges/splits possible; stopping.")
        break

    final_sizes = [len(s) for s in fragments]

    if verbose and not all(min_size <= s <= max_size for s in final_sizes):
        print(
            f"Warning: some fragments still outside {min_size}-{max_size}: "
            f"{final_sizes}"
        )

    return fragments


def extract_fragment(mol: Chem.Mol, atom_indices: Iterable[int]) -> Chem.Mol:
    """
    Extract a Fragment of a Molecule object as a Chem.Mol object based on selected atom indexes
    """
    keep = sorted(set(int(i) for i in atom_indices))

    n = mol.GetNumAtoms()
    if keep[0] < 0 or keep[-1] > n:
        raise IndexError(f"atom index out of range (0..{n - 1}): {keep}")

    keep_set = set(keep)

    # Map old atom idx -> new atom idx
    rw = Chem.RWMol()
    old2new = {}
    for old_i in keep:
        # Copy the atom object to preserve charge/isotope/aromaticity/etc.
        new_i = rw.AddAtom(Chem.Atom(mol.GetAtomWithIdx(old_i)))
        old2new[old_i] = new_i

    # Add bonds that connect kept atoms
    for b in mol.GetBonds():
        a, c = b.GetBeginAtomIdx(), b.GetEndAtomIdx()
        if a in keep_set and c in keep_set:
            rw.AddBond(old2new[a], old2new[c], b.GetBondType())
            nb = rw.GetBondBetweenAtoms(old2new[a], old2new[c])
            nb.SetIsAromatic(b.GetIsAromatic())

    out = rw.GetMol()

    # Copy coordinates for EACH conformer
    out.RemoveAllConformers()
    for conf in mol.GetConformers():
        new_conf = Chem.Conformer(len(keep))
        new_conf.Set3D(conf.Is3D())
        for old_i in keep:
            new_conf.SetAtomPosition(old2new[old_i], conf.GetAtomPosition(old_i))
        out.AddConformer(new_conf, assignId=True)

    return out



import torch
from rdkit import Chem

from .cheminformatics.pipeline import set_conformer_positions
from .cheminformatics.shape_similarity import best_pi_rotation_by_tanimoto
from .conformer_generator import MLConformerGenerator
from .utils import (MAX_FRAG_SIZE, MIN_FRAG_SIZE, align_mol_to_principal_frame,
                    apply_transform, concat_masked_and_pad, extract_fragment,
                    get_context_shape, ifm_get_xh_from_fragment,
                    ifm_prepare_fragments_for_merge,
                    ifm_prepare_gen_fragment_context, inverse_coord_transform,
                    prepare_edm_input, samples_to_rdkit_mol,
                    split_molecule_size_constrained, standardize_mol)


def inertial_fragment_matching(
    reference_conformer: Chem.Mol,
    n_samples: int,
    generator: MLConformerGenerator,
    merger: MLConformerGenerator = None,
    variance: int = 1,
    n_atoms: int = None,
    resample_steps: int = 0,
    diffusion_steps_merging: int = 10,
    min_frag_size: int = MIN_FRAG_SIZE,
    max_frag_size: int = MAX_FRAG_SIZE,
    max_iter: int = 200,
    verbose: bool = False,
    predict_bonds: bool = False,
    optimize_geometry: bool = False,
) -> list[Chem.Mol]:
    """
    Inference pipeline for Inertial Fragment Matching.

    The pipeline attempts to split the reference molecule into fragments by cutting
    single non-ring bonds. Fragments within the specified size range are selected.
    For each selected fragment, shape-similar fragments are generated using a
    generator model. These generated fragments replace the corresponding reference
    fragments and are subsequently combined using a merger model to produce full
    molecule candidates.

    :param reference_conformer: Reference molecule as an RDKit Mol object.
    :param n_samples: Number of molecules to generate.
    :param generator: Model used to generate initial fragments
    :param merger: Model used to merge generated fragments
    :param variance: Allowed deviation (± heavy atoms) from the requested n_atoms in generated molecules.
    :param n_atoms: Target median number of heavy atoms in generated molecules.
                    If None, the number of heavy atoms in the reference molecule is used.
    :param resample_steps: Number of harmonizing diffusion resampling steps.
    :param diffusion_steps_merging: Number of diffusion steps used during fragment
                                    merging. A typical value is ~10% of the total diffusion steps of the merger model.
    :param min_frag_size: Minimum allowed fragment size when splitting the molecule.
    :param max_frag_size: Maximum allowed fragment size when splitting the molecule.
    :param max_iter: Maximum number of attempts to split the reference molecule.
    :param verbose: If True, enables additional logging.
    :param predict_bonds: If True, bonds are predicted using AdjMatSeer.
    :param optimize_geometry: If True, performs standard geometry optimization on the generated molecules.
    :return: List of generated molecules as RDKit Mol objects.
    """

    if merger is None:
        merger = generator

    g_device = generator.device
    m_device = merger.device

    g_context_norms = generator.context_norms
    m_context_norms = merger.context_norms

    # Strip of Hs and align Reference to principal Inertial Frame, saving rotation and shift
    ref_mol = Chem.RemoveHs(reference_conformer)
    ref_context, shift, rotation, aligned_ref_coord = align_mol_to_principal_frame(
        ref_mol
    )

    if n_atoms is None:
        n_nodes = aligned_ref_coord.size(0)
    else:
        n_nodes = n_atoms

    min_n_atoms_final = n_nodes - variance
    max_n_atoms_final = n_nodes + variance

    aligned_ref_mol = set_conformer_positions(ref_mol, aligned_ref_coord)

    # Split Reference molecule into fragments
    fragment_sets = split_molecule_size_constrained(
        mol=aligned_ref_mol,
        min_size=min_frag_size,
        max_size=max_frag_size,
        max_iter=max_iter,
        verbose=verbose,
    )

    if len(fragment_sets) == 0:
        raise RuntimeError(
            "Could not split reference molecule into fragments, aborting IFM generation."
        )

    # Extract fragments as individual conformers
    extracted_frags = []

    for frag_set in fragment_sets:
        extracted_frags.append(extract_fragment(aligned_ref_mol, frag_set))

    fragment_contexts = []
    fragment_shifts = []
    fragment_rotations = []
    ref_fragment_coords = []

    edm_inputs = []

    # Determine the max number of nodes
    max_n_nodes = 0
    for frag in extracted_frags:
        _n_atoms = frag.GetNumHeavyAtoms()
        if _n_atoms > max_n_nodes:
            max_n_nodes = _n_atoms

    # Align Fragments to their respective Principal Inertial Frames,
    # while remembering corresponding Shifts and Rotations
    # Prepare concatenate-able edm inputs for all fragments
    for frag in extracted_frags:
        f_n_atoms = frag.GetNumHeavyAtoms()
        f_context, f_shift, f_rotation, f_coord = align_mol_to_principal_frame(frag)
        fragment_contexts.append(f_context)
        fragment_shifts.append(f_shift)
        fragment_rotations.append(f_rotation)
        ref_fragment_coords.append(f_coord)

        node_mask, edge_mask, batch_context = prepare_edm_input(
            n_samples=n_samples,
            reference_context=f_context,
            context_norms=g_context_norms,
            min_n_nodes=f_n_atoms,
            max_n_nodes=f_n_atoms,
            device=g_device,
            pad_to=max_n_nodes,
        )

        edm_inputs.append(
            {
                "node_mask": node_mask,
                "edge_mask": edge_mask,
                "batch_context": batch_context,
            }
        )

    total_node_mask = torch.cat([x["node_mask"] for x in edm_inputs], 0)
    total_batched_context = torch.cat([x["batch_context"] for x in edm_inputs])

    # Total Edge mask is a bit trickier:
    helper_node_mask = total_node_mask.clone().squeeze()
    batch_size = helper_node_mask.size(0)
    max_n_nodes = helper_node_mask.size(1)

    # Compute Total Edge Mask
    total_edge_mask = helper_node_mask.unsqueeze(1) * helper_node_mask.unsqueeze(2)
    diag_mask = ~torch.eye(
        total_edge_mask.size(1), dtype=torch.bool, device=g_device
    ).unsqueeze(0)
    total_edge_mask *= diag_mask
    total_edge_mask = total_edge_mask.view(batch_size * max_n_nodes * max_n_nodes, 1)

    # Generating samples matching all isolated fragments at the same time
    with torch.no_grad():
        total_x, total_h = generator.generative_model(
            total_node_mask,
            total_edge_mask,
            total_batched_context,
            resample_steps=resample_steps,
        )

    # Split generated tensor into samples generated per fragment
    x_fragments = torch.split(total_x, n_samples, dim=0)
    h_fragments = torch.split(total_h, n_samples, dim=0)
    node_masks = torch.split(total_node_mask, n_samples, 0)

    # Batchify shifts and rotations
    # Apply corresponding inverse coord transforms to each fragment coordinates
    coord_for_merge = []

    for i, frag_coord in enumerate(x_fragments):
        batch_shift = fragment_shifts[i].unsqueeze(0).expand(n_samples, -1).to(m_device)
        batch_rot = (
            fragment_rotations[i]
            .transpose(0, 1)
            .unsqueeze(0)
            .expand(n_samples, 3, -1)
            .to(m_device)
        )

        # Align generated fragments to principal frame to maximize ref fragment volume overlay
        aligned_x = []
        frag_coord = frag_coord.to("cpu")
        for old_x in frag_coord:
            try:
                aligned_x.append(align_coord(cand_coord=old_x, ref_coord=ref_fragment_coords[i]))
            except RuntimeError:
                pass

        aligned_x = torch.stack(aligned_x, dim=0).to(m_device)

        new_x = inverse_coord_transform(
            coord=aligned_x,
            shift=batch_shift,
            rotation=batch_rot.transpose(1, 2),
        )

        # Save prepared coordinates for Merging the fragments
        coord_for_merge.append(new_x)

    # We merge all the fragments by dropping masked atoms to get a proper z_seed
    merged_x = concat_masked_and_pad(
        coord_for_merge, node_masks, pad_to=max_n_atoms_final
    )
    merged_h = concat_masked_and_pad(h_fragments, node_masks, pad_to=max_n_atoms_final)

    z_seed = torch.cat([merged_x, merged_h], dim=2).to(m_device)

    # Here we prepare masks as for normal generation
    merging_node_mask, merging_edge_mask, batch_ref_context = prepare_edm_input(
        n_samples=n_samples,
        reference_context=ref_context,
        context_norms=m_context_norms,
        min_n_nodes=min_n_atoms_final,
        max_n_nodes=max_n_atoms_final,
        device=m_device,
    )

    with torch.no_grad():
        final_x, final_h = merger.generative_model.ifm_merge_fragments(
            node_mask=merging_node_mask,
            edge_mask=merging_edge_mask,
            context=batch_ref_context,
            z_seed=z_seed,
            diffusion_level=diffusion_steps_merging,
            resample_steps=resample_steps,
        )

    ifm_mols = samples_to_rdkit_mol(
        positions=final_x,
        one_hot=final_h,
        node_mask=merging_node_mask,
        atom_decoder=merger.atom_decoder,
    )

    if predict_bonds:
        raw_mols = merger.predict_bonds(ifm_mols)
        ifm_mols = []
        for f_mol in raw_mols:
            std_mol = standardize_mol(
                mol=f_mol, optimize_geometry=optimize_geometry, ifm_mode=True
            )
            if std_mol:
                ifm_mols.append(std_mol)

    return ifm_mols


# FIXED FRAGMENT GENERATION
# ------------------------------------------------------


def ff_inertial_fragment_matching(
    fixed_fragment: Chem.Mol | set,
    n_samples: int,
    generator: MLConformerGenerator,
    reference_conformer: Chem.Mol = None,
    reference_context: torch.Tensor = None,
    n_atoms: int = None,
    merger: MLConformerGenerator = None,
    variance: int = 1,
    resample_steps: int = 0,
    blend_power: int = 3,
    diffusion_steps_merging: int = 10,
    predict_bonds: bool = False,
    optimize_geometry: bool = False,
) -> list[Chem.Mol]:
    """
    Inference pipeline for Inertial Fragment Matching with a fixed fragment.

    The pipeline uses the moments-of-inertia (MOI) tensors of the reference
    molecule and the fixed fragment to compute the residual tensor corresponding
    to the fragment that needs to be generated. This residual MOI tensor is then
    used to guide fragment generation with the generator model.

    The generated fragments are subsequently merged with the fixed fragment using
    a merger model, while preserving and injecting the fixed fragment into the
    resulting molecule.

    :param fixed_fragment: a fragment to be fixed as a RDkit Mol object or a set of indexes of atoms in the reference molecule
    :param n_samples: Number of molecules to generate.
    :param generator: Model used to generate initial fragments
    :param reference_conformer: Reference molecule as an RDKit Mol object.
    :param reference_context: Arbitrary Reference context if applicable, instead of reference_conformer
    :param n_atoms: Target median number of heavy atoms in generated molecules.
                    If None, the number of heavy atoms in the reference molecule is used.
                    Required if using reference_context
    :param merger: Model used to merge generated fragments
    :param variance: Allowed deviation (± heavy atoms) from the requested n_atoms in generated molecules.
    :param resample_steps: Number of harmonizing diffusion resampling steps.
    :param blend_power: power of the polynomial blending schedule
    :param diffusion_steps_merging: Number of diffusion steps used during fragment
                                    merging. A typical value is ~10% of the total diffusion steps of the merger model.
    :param predict_bonds: If True, bonds are predicted using AdjMatSeer.
    :param optimize_geometry: If True, performs standard geometry optimization on the generated molecules.
    :return: List of generated molecules as RDKit Mol objects.
    """

    if merger is None:
        merger = generator

    g_context_norms = generator.context_norms
    m_context_norms = merger.context_norms

    g_device = generator.device
    m_device = merger.device

    # Strip of Hs and align Reference to principal Inertial Frame, saving rotation and shift

    if reference_conformer:
        ref_conformer = Chem.RemoveHs(reference_conformer)
        ref_context, shift, rotation, aligned_ref_coord = align_mol_to_principal_frame(
            ref_conformer
        )

        aligned_ref_mol = set_conformer_positions(ref_conformer, aligned_ref_coord)

        if isinstance(fixed_fragment, set):
            ref_idx = {atom.GetIdx() for atom in ref_conformer.GetAtoms()}
            ref_frag_idx = ref_idx - fixed_fragment
            fixed_fragment = extract_fragment(aligned_ref_mol, fixed_fragment)
            ref_fragment = extract_fragment(aligned_ref_mol, ref_frag_idx)
            # Align ref fragment
            _, _, _, ref_fragment_coords = align_mol_to_principal_frame(ref_fragment)

            def _alignment_func(a):
                return align_coord(cand_coord=a, ref_coord=ref_fragment_coords)

        elif isinstance(fixed_fragment, Chem.Mol):
            # Apply the Reference Transformation to Fixed fragment to keep consistency
            fixed_fragment = Chem.RemoveAllHs(fixed_fragment)
            ff_conf = fixed_fragment.GetConformer()
            ff_coord = torch.tensor(ff_conf.GetPositions(), dtype=torch.float32)
            ff_coord_ref_aligned = apply_transform(ff_coord, shift, rotation)
            fixed_fragment = set_conformer_positions(
                fixed_fragment, ff_coord_ref_aligned
            )

            def _alignment_func(coord):
                return align_coord(cand_coord=coord, ref_coord=None)

        else:
            raise ValueError(
                f"Unsupported fixed fragment type - {type(fixed_fragment)}"
            )

        n_nodes = aligned_ref_coord.size(0)
        min_n_nodes = n_nodes - variance
        max_n_nodes = n_nodes + variance

    elif reference_context is not None:
        if n_atoms is None:
            raise ValueError(
                "Reference Number of Atoms should be provided, when generating samples using context."
            )

        if isinstance(fixed_fragment, Chem.Mol):
            fixed_fragment = Chem.RemoveAllHs(fixed_fragment)

            def _alignment_func(coord):
                return align_coord(cand_coord=coord, ref_coord=None)

        else:
            raise ValueError(
                "'fixed_fragment' must be a Mol object when generating from a reference context."
            )

        min_n_nodes = n_atoms - variance
        max_n_nodes = n_atoms + variance
        ref_context = reference_context

    else:
        raise ValueError(
            "Either a reference RDkit Mol object or context as torch.Tensor should be provided for generation."
        )

    ff_coord = torch.tensor(
        fixed_fragment.GetConformer().GetPositions(), dtype=torch.float32
    ).to(g_device)

    node_mask, edge_mask, batch_context = prepare_edm_input(
        n_samples=n_samples,
        reference_context=ref_context,
        context_norms=m_context_norms,
        min_n_nodes=min_n_nodes,
        max_n_nodes=max_n_nodes,
        device=g_device,
    )

    ff_n_nodes = torch.sum(node_mask, dim=1).to(torch.long)

    (
        frag_node_mask,
        frag_edge_mask,
        batched_normed_frag_context,
        shift,
        rotation,
    ) = ifm_prepare_gen_fragment_context(
        fixed_fragment_x=ff_coord,
        reference_context=ref_context,
        context_norms=g_context_norms,
        n_nodes=ff_n_nodes,
        max_n_nodes=max_n_nodes,
        min_n_nodes=min_n_nodes,
        device=g_device,
    )

    with torch.no_grad():
        total_x, total_h = generator.generative_model(
            frag_node_mask,
            frag_edge_mask,
            batched_normed_frag_context,
            resample_steps=resample_steps,
        )

    # Aligning generated fragments to corresponding places inside the reference molecule

    aligned_x = []
    frag_coord = total_x.to("cpu")

    for old_x in frag_coord:
        try:
            aligned_x.append(_alignment_func(old_x))
        except RuntimeError:
            pass

    aligned_x = torch.stack(aligned_x, dim=0).to(m_device)

    new_x = inverse_coord_transform(
        coord=aligned_x,
        shift=shift.to(m_device),
        rotation=rotation.to(m_device),
    )

    fixed_fragment_x, fixed_fragment_h = ifm_get_xh_from_fragment(
        fixed_fragment=fixed_fragment, device=m_device
    )

    z_known, fixed_mask = ifm_prepare_fragments_for_merge(
        fixed_fragment_x=fixed_fragment_x,
        fixed_fragment_h=fixed_fragment_h,
        gen_fragments_x=new_x,
        gen_fragments_h=total_h.to(m_device),
        device=m_device,
        max_n_nodes=max_n_nodes,
    )

    with torch.no_grad():
        final_x, final_h = merger.generative_model.ifm_merge_fragments_with_injection(
            node_mask.to(m_device),
            edge_mask.to(m_device),
            fixed_mask,
            context=batch_context.to(m_device),
            z_seed=z_known,
            diffusion_level=diffusion_steps_merging,
            resample_steps=resample_steps,
            blend_power=blend_power,
        )

    ff_ifm_mols = samples_to_rdkit_mol(
        positions=final_x,
        one_hot=final_h,
        node_mask=node_mask,
        atom_decoder=merger.atom_decoder,
    )

    if predict_bonds:
        raw_mols = merger.predict_bonds(ff_ifm_mols)
        ff_ifm_mols = []
        for f_mol in raw_mols:
            std_mol = standardize_mol(
                mol=f_mol, optimize_geometry=optimize_geometry, ifm_mode=True
            )
            if std_mol:
                ff_ifm_mols.append(std_mol)

    return ff_ifm_mols


def align_coord(
    cand_coord: torch.Tensor, ref_coord: torch.Tensor = None
) -> torch.Tensor:
    # move coord to center
    virtual_com = torch.mean(cand_coord, dim=0)
    cand_coord = cand_coord - virtual_com

    # Get Coords in Principal Frame
    _, aligned_coord = get_context_shape(cand_coord, include_rotation=False)

    if ref_coord is not None:
        ref_coord = ref_coord - virtual_com
        best_coord, _ = best_pi_rotation_by_tanimoto(ref_coord, aligned_coord)
        return best_coord

    return aligned_coord



from typing import List, Tuple

import numpy as np
from rdkit import Chem

from ..utils.common import bond_type_dict, canonicalise
from ..utils.config import DIMENSION
from .molgraph import MolGraphONNX


def prepare_masks_onnx(
    n_nodes: np.ndarray,
    max_n_nodes: int,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Prepare node and edge masks torch-free
    :param n_nodes: a list of sizes of requested molecules (n_samples, 1)
    :param max_n_nodes: maximal number of nodes
    :return: node_mask, edge_mask
    """

    batch_size = n_nodes.shape[0]

    node_mask = np.zeros((batch_size, max_n_nodes))
    for i in range(batch_size):
        node_mask[i, 0 : n_nodes[i]] = 1

    # Compute edge_mask
    edge_mask = np.expand_dims(node_mask, 1) * np.expand_dims(node_mask, 2)
    diag_mask = ~np.eye(edge_mask.shape[1], dtype=bool)[None, :, :]
    edge_mask *= diag_mask
    edge_mask = edge_mask.reshape(-1, 1)
    node_mask = np.expand_dims(node_mask, 2)

    return node_mask, edge_mask


def prepare_edm_input_onnx(
    n_samples: int,
    reference_context: np.ndarray,
    context_norms: dict,
    min_n_nodes: int,
    max_n_nodes: int,
    pad_to: int = None,
):
    pad_dim = pad_to if pad_to is not None else max_n_nodes

    # Create a random list of sizes between min_n_nodes and max_n_nodes of length n_samples
    nodesxsample = np.random.randint(min_n_nodes, max_n_nodes + 1, size=n_samples)

    batch_size = nodesxsample.shape[0]

    node_mask, edge_mask = prepare_masks_onnx(n_nodes=nodesxsample, max_n_nodes=pad_dim)

    normed_context = (reference_context - context_norms["mean"]) / context_norms["mad"]

    batch_context = np.repeat(normed_context[None, :], batch_size, axis=0)

    batch_context = np.repeat(batch_context[:, None, :], pad_dim, axis=1) * node_mask

    return (
        node_mask,
        edge_mask,
        batch_context,
    )


def samples_to_rdkit_mol_onnx(
    positions,
    one_hot: np.ndarray,
    node_mask: np.ndarray = None,
    atom_decoder: dict = None,
) -> List[Chem.Mol]:
    """
    Convert EDM Samples to RDKit mol objects torch-free
    :param positions: coordinates tensor
    :param one_hot: one-hot encoded atom types
    :param node_mask: node mask
    :param atom_decoder: atom decoder dictionary
    :return: a list of samples as RDKit Mol objects without bond information
    """
    rdkit_mols = []

    if node_mask is not None:
        atomsxmol = np.sum(node_mask, axis=1).flatten()
    else:
        atomsxmol = [one_hot.shape[1]] * one_hot.shape[0]

    for batch_i in range(one_hot.shape[0]):
        xyz_block = "%d\n\n" % atomsxmol[batch_i]
        atoms = np.argmax(one_hot[batch_i], axis=1)
        n_atoms = int(atomsxmol[batch_i])
        for atom_i in range(n_atoms):
            atom = atoms[atom_i]
            atom = atom_decoder[atom.item()]
            xyz_block += "%s %.9f %.9f %.9f\n" % (
                atom,
                positions[batch_i, atom_i, 0],
                positions[batch_i, atom_i, 1],
                positions[batch_i, atom_i, 2],
            )

        mol = Chem.MolFromXYZBlock(xyz_block)
        if mol is not None:
            rdkit_mols.append(mol)

    return rdkit_mols


def prepare_adj_mat_seer_input_onnx(
    mols: List[Chem.Mol],
    dimension: int,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[Chem.Mol]]:
    canonicalised_samples = []
    n_samples = len(mols)

    elements_batch = np.zeros((n_samples, dimension), dtype=np.int64)
    dist_mat_batch = np.zeros((n_samples, dimension, dimension), dtype=np.float32)
    adj_mat_batch = np.zeros((n_samples, dimension, dimension), dtype=np.float32)

    for i, sample in enumerate(mols):
        mol = canonicalise(sample)

        conf = mol.GetConformer()
        coord = np.array(conf.GetPositions(), dtype=np.float32)

        structure = MolGraphONNX.from_mol(mol=mol, remove_hs=False)
        elements = structure.elements_vector()
        n_atoms = np.count_nonzero(elements)

        target_adjacency_matrix = structure.adjacency_matrix()

        sc_adj_mat = np.argmax(target_adjacency_matrix, axis=2) + np.eye(dimension)

        sc_adj_mat[sc_adj_mat > 0] = 1

        dist_mat = distance_matrix(coord)
        pad_width = ((0, dimension - n_atoms), (0, dimension - n_atoms))
        pad_dist_mat_sc = np.pad(
            dist_mat, pad_width, mode="constant", constant_values=0
        )
        pad_dist_mat_sc += np.eye(dimension)

        elements_batch[i] = elements.astype(np.int64)
        dist_mat_batch[i] = pad_dist_mat_sc
        adj_mat_batch[i] = sc_adj_mat
        canonicalised_samples.append(mol)

    return elements_batch, dist_mat_batch, adj_mat_batch, canonicalised_samples


def prepare_fragment_onnx(
    n_samples: int,
    fragment: Chem.Mol,
    max_n_nodes: int = DIMENSION,
    min_n_nodes: int = 15,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Torch-free preparation of a Fixed Fragment for Inpainting. Converts Mol to latent Z tensor, ready for injection
    :param n_samples: required batch size of the prepared latent fragment - number of molecules to generate
    :param fragment: fragment to prepare rdkit Mol
    :param max_n_nodes: possible maximum number of nodes - for padding - int
    :param min_n_nodes: possible minimum number of nodes - int
    :return: Latent representation of the fragment and a mask,
             indicating which atoms in the latent representation are fixed
    """

    # Remove Hs
    coord, h = ifm_get_xh_from_fragment_onnx(fragment)
    n_atoms = coord.shape[0]

    # Check that fragment size is adequate
    if n_atoms >= min_n_nodes:
        raise ValueError(
            "Fragment must contain fewer atoms than minimum generation size."
        )
    if n_atoms >= max_n_nodes:
        raise ValueError(
            "Fragment has more atoms than the maximum number of atoms requested."
        )

    x = np.pad(coord, ((0, max_n_nodes - n_atoms), (0, 0)), mode="constant")
    h = np.pad(h, ((0, max_n_nodes - n_atoms), (0, 0)), mode="constant")

    # Batch x and h

    x = np.tile(x[None, :, :], (n_samples, 1, 1))  # (n_samples, max_n_nodes, 3)
    h = np.tile(h[None, :, :], (n_samples, 1, 1))
    z_known = np.concatenate([x, h], axis=2)

    fixed_mask = np.zeros((n_samples, max_n_nodes, 1), dtype=np.float32)
    fixed_mask[:, :n_atoms, 0] = 1.0

    return z_known, fixed_mask


def redefine_bonds_onnx(mol: Chem.Mol, adj_mat: np.ndarray) -> Chem.Mol:
    n = mol.GetNumAtoms()
    # Pass the molecule through xyz block to remove bonds and all extra atom properties
    i_xyz = Chem.MolToXYZBlock(mol)
    c_mol = Chem.MolFromXYZBlock(i_xyz)
    ed_mol = Chem.EditableMol(c_mol)

    repr_m = np.tril(np.argmax(adj_mat, axis=2))
    repr_m = repr_m * (1 - np.eye(repr_m.shape[0], dtype=int))

    for i in range(n):
        for j in range(n):
            # Find out the bond type by indexing 1 in the matrix bond
            bond_type = repr_m[i, j].item()

            if bond_type != 0:
                ed_mol.AddBond(i, j, bond_type_dict[bond_type])

    new_mol = ed_mol.GetMol()

    return new_mol


def distance_matrix(coordinates: np.ndarray) -> np.ndarray:
    """
    Generates a distance matrices from a xyz coordinates tensor
    :param coordinates: xyz coordinates tensor
    :return: distance matrix
    """
    n = coordinates.shape[0]
    i_mat = np.repeat(coordinates[np.newaxis, :, :], n, axis=0)
    # Repeat coordinates tensor along new dimension
    j_mat = np.transpose(i_mat, (1, 0, 2))

    dist_matrix = np.sqrt(np.sum((i_mat - j_mat) ** 2, axis=2))

    return dist_matrix


def get_context_shape_onnx(
    coord: np.ndarray, include_rotation: bool = False
) -> Tuple[np.ndarray, np.ndarray] | Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Finds the principal axes for the conformer,
    and calculates Moment of Inertia tensor for the conformer in principal axes.
    All atom masses are considered equal to one, to capture shape only.
    :param coord: initial coordinates of the atoms
    :param include_rotation: bool flag, if True rotation (eigenvectors) applied to coord is returned
    :return: Principal components of MOI tensor, and coordinates rotated to a principal frame as a tuple of ndarrays
    """
    masses = np.ones(coord.shape[0])
    moi_tensor = get_moment_of_inertia_tensor_onnx(coord, masses)
    # Diagonalize the MOI tensor using eigen decomposition
    _, eigenvectors = np.linalg.eigh(moi_tensor)

    # Rotate points to principal axes
    rotated_points = np.matmul(coord.astype(np.float32), eigenvectors)

    # Get the three main moments of inertia from the main diagonal
    context = np.diag(get_moment_of_inertia_tensor_onnx(rotated_points, masses))

    if include_rotation:
        return context, rotated_points, eigenvectors
    else:
        return context, rotated_points


def get_moment_of_inertia_tensor_onnx(
    coord: np.ndarray, weights: np.ndarray
) -> np.ndarray:
    """
    Calculate a Moment of Inertia tensor
    :return: Moment of Inertia Tensor in input coordinates
    """
    x, y, z = coord[:, 0], coord[:, 1], coord[:, 2]

    # Diagonal elements
    i_xx = np.sum(weights * (y**2 + z**2))
    i_yy = np.sum(weights * (x**2 + z**2))
    i_zz = np.sum(weights * (x**2 + y**2))

    # Off-diagonal elements
    i_xy = -np.sum(x * y)
    i_xz = -np.sum(x * z)
    i_yz = -np.sum(y * z)

    # Construct the MOI tensor
    moi_tensor = np.array(
        [[i_xx, i_xy, i_xz], [i_xy, i_yy, i_yz], [i_xz, i_yz, i_zz]],
        dtype=np.float32,
    )

    return moi_tensor


def ifm_get_xh_from_fragment_onnx(
    fixed_fragment: Chem.Mol,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Get coordinates and atom types as tensors for a fragment torch-free
    :param fixed_fragment: fragment as rdkit Mol object
    :return: fixed_fragment_x, fixed_fragment_h
    """
    # Get coordinates of the fixed fragment
    ff_mol = Chem.RemoveAllHs(fixed_fragment)
    ff_conformer = ff_mol.GetConformer()
    ff_x = np.array(ff_conformer.GetPositions(), dtype=np.float32)

    # Get atom types of a fixed fragment
    ff_structure = MolGraphONNX.from_mol(mol=ff_mol, remove_hs=True)
    ff_n_atoms = ff_x.shape[0]

    ff_h = ff_structure.one_hot_elements_encoding(
        ff_n_atoms
    )  # Atom types of a fixed fragment

    return ff_x, ff_h


def ifm_prepare_gen_fragment_context_onnx(
    fixed_fragment_x: np.ndarray,
    reference_context: np.ndarray,
    context_norms: dict,
    n_nodes: np.ndarray,
    max_n_nodes: int,
    min_n_nodes: int,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Prepare Contexts for Generation of Individual Fragments based on Reference and Fixed Fragment MOI tensors.
    Torch-free.
    Fixed Fragment should be expressed in the same coordinate system as the reference, The origin of such coordinate
    system should be place in the COM of the Reference.

    :param fixed_fragment_x:  coordinates of the fixed fragment atoms
    :param reference_context: Principal components of MOI Tensor of the reference shape (3)
    :param context_norms: Values to norm the context for individual fragment generation
    :param n_nodes: a list of sizes of requested molecules (n_samples, 1)
    :param max_n_nodes: a maximal number of nodes allowed for the whole generated molecule
    :param min_n_nodes: a minimal number of nodes allowed for the whole generated molecule
    :return: frag_node_mask, frag_edge_mask, batched_normed_frag_context, shift, rotation
    """
    batch_size = n_nodes.shape[0]

    ff_n_atoms = fixed_fragment_x.shape[0]
    masses_ff = np.ones(ff_n_atoms)

    if ff_n_atoms >= min_n_nodes:
        raise ValueError(
            "Fragment must contain fewer atoms than minimum generation size."
        )
    if ff_n_atoms >= max_n_nodes:
        raise ValueError(
            "Fragment has more atoms than the maximum number of atoms requested."
        )

    # Fixed fragment MOI around origin
    moi_ff = get_moment_of_inertia_tensor_onnx(fixed_fragment_x, masses_ff)  # (3, 3)

    # Reference MOI as diagonal matrix (expand to B)
    moi_ref = np.diag(reference_context)  # (3, 3)

    moi_ref_batch = np.tile(moi_ref[None, :, :], (batch_size, 1, 1))  # (B, 3, 3)
    moi_ff_batch = np.tile(moi_ff[None, :, :], (batch_size, 1, 1))  # (B, 3, 3)

    # Gen frag MOI around origin
    moi_gen_origin = moi_ref_batch - moi_ff_batch  # (B, 3, 3)

    # COM of fixed fragment
    com_ff = fixed_fragment_x.mean(axis=0)  # (3,)

    gen_n_atoms = n_nodes.reshape(batch_size, 1).astype(float) - ff_n_atoms  # (B, 1)

    # COM of generated fragments in respect to number of atoms in each
    shift = (ff_n_atoms * com_ff.reshape(1, 3)) / gen_n_atoms  # (B, 3)

    # Shift MOI to COM of generated fragment
    moi_gen_com = shift_moi_to_com_batch_onnx(
        moi_gen_origin, shift, gen_n_atoms
    )  # (B, 3, 3)

    # Diagonalize MOI of generated fragment
    frag_context, rotation = np.linalg.eigh(moi_gen_com)  # (B, 3), (B, 3, 3)
    normed_frag_context = (frag_context - context_norms["mean"]) / context_norms["mad"]

    max_n_nodes_frag = max_n_nodes - ff_n_atoms

    # Flatten gen_n_atoms and convert to int
    gen_n_atoms = gen_n_atoms.reshape(-1).astype(np.int64)

    frag_node_mask, frag_edge_mask = prepare_masks_onnx(
        n_nodes=gen_n_atoms,
        max_n_nodes=max_n_nodes_frag,
    )

    batched_normed_frag_context = (
        np.repeat(normed_frag_context[:, None, :], max_n_nodes_frag, axis=1)
        * frag_node_mask
    )

    return frag_node_mask, frag_edge_mask, batched_normed_frag_context, shift, rotation


def ifm_prepare_fragments_for_merge_onnx(
    fixed_fragment_x: np.ndarray,
    fixed_fragment_h: np.ndarray,
    gen_fragments_x: np.ndarray,
    gen_fragments_h: np.ndarray,
    max_n_nodes: int,
):
    """
    Prepares Multiple Fragments for Merge. Prepares latent Z tensor, ready for injection and mask for a fixed fragment.
    Torch-free
    :param fixed_fragment_x: Coordinates of a fixed fragment as np.ndarray
    :param fixed_fragment_h: One-hot encoded atom types of a fixed fragment as np.ndarray
    :param gen_fragments_x: Batch of coordinates of generated fragments - np.ndarray
    :param gen_fragments_h: Batch of One-hot encoded atom types of generated fragments - np.ndarray
    :param max_n_nodes: maximal allowable number of atoms
    :return: z_known, fixed_mask, n_samples
    """

    n_samples = gen_fragments_x.shape[0]

    ff_n_atoms = fixed_fragment_x.shape[0]

    # Add a batch dimension
    ff_x = fixed_fragment_x[None, :, :]  # Shape: (1, N, 3)
    ff_h = fixed_fragment_h[None, :, :]  # Shape: (1, N, F)

    # Repeat across batch dimension
    ff_x_batched = np.repeat(ff_x, n_samples, axis=0)  # Shape: (n_samples, N, 3)
    ff_h_batched = np.repeat(ff_h, n_samples, axis=0)

    x_prep = np.concatenate([ff_x_batched, gen_fragments_x], axis=1)
    h_prep = np.concatenate([ff_h_batched, gen_fragments_h], axis=1)

    z_known = np.concatenate([x_prep, h_prep], axis=2)

    # The fixed fragment is always in the first place - so we set fixed mask to have 1s only
    # on the first ff_n_atoms elements of z_known

    fixed_mask = np.zeros((n_samples, max_n_nodes, 1), dtype=np.float32)
    fixed_mask[:, :ff_n_atoms, 0] = 1.0

    return z_known, fixed_mask


def shift_moi_to_com_batch_onnx(
    moi_origin: np.ndarray, r_coms: np.ndarray, masses: np.ndarray
) -> np.ndarray:
    """
    Translates moment of inertia from the origin to multiple guessed centers of mass
    using the inverse parallel axis theorem. Torch-Free

    :param moi_origin: (3, 3) Inertia tensor around origin (shared across batch)
    :param r_coms: (B, 3) Vectors from origin to guessed COMs
    :param masses: (B,) Total masses per example
    :return: I_coms: (B, 3, 3) Inertia tensors about the guessed COMs
    """
    batch_size = r_coms.shape[0]
    i_3 = np.tile(np.eye(3)[None, :, :], (batch_size, 1, 1))  # (B, 3, 3)

    r = r_coms.reshape(batch_size, 3, 1)  # (B, 3, 1)
    r_outer = np.matmul(r, np.transpose(r, (0, 2, 1)))  # (B, 3, 3)
    r_norm_sq = np.sum(r_coms**2, axis=1).reshape(batch_size, 1, 1)  # (B, 1, 1)

    masses = masses.reshape(batch_size, 1, 1)  # (B, 1, 1)
    shift = masses * (r_norm_sq * i_3 - r_outer)  # (B, 3, 3)

    return moi_origin - shift  # (B, 3, 3)


def inverse_coord_transform_onnx(
    coord: np.ndarray, shift: np.ndarray, rotation: np.ndarray
) -> np.ndarray:
    """
    Inverse shift and Rotation transformation to a batch of xyz coordinates sets. Torch-free
    :param coord: Batch of Coordinates to be modified (batch_size, N, 3)
    :param shift: Batch of shifts to be applied (batch_size, 3)
    :param rotation: Batch of Rotations to be applied (batch_size, 3, 3)
    :return: Modified Coordinates
    """
    # Rotate first
    batch_size = coord.shape[0]
    x_rotated = np.matmul(coord, np.transpose(rotation, (0, 2, 1)))
    # Translate second
    x_translated = x_rotated - shift.reshape(batch_size, 1, 3)

    return x_translated


def coord_to_pf_batched_onnx(coord: np.ndarray) -> np.ndarray:
    """
    Puts a batch of coordinates to their principal inertial frame
    :param coord: (B, N, 3) coordinates batched
    :return: (B, N, 3) coordinates in pf batched
    """
    # (N,) weights like torch.ones(coord.shape[-2])
    weights = np.ones(coord.shape[-2], dtype=np.float32)

    coord_b_f = coord.astype(np.float32, copy=False)  # (B, N, 3)
    com_b = coord_b_f.mean(axis=1, keepdims=True)  # (B, 1, 3)
    coord_b_f = coord_b_f - com_b  # center coordinates

    # must return (B, 3, 3)
    moi = get_moment_of_inertia_tensor_batched_onnx(coord_b_f, weights)

    # NumPy: eigh returns (eigenvalues, eigenvectors), eigenvectors are columns
    _, eigenvectors = np.linalg.eigh(moi)  # (B, 3, 3)

    # Batched matmul: (B, N, 3) @ (B, 3, 3) -> (B, N, 3)
    rotated = coord_b_f @ eigenvectors

    return rotated


def get_moment_of_inertia_tensor_batched_onnx(
    coord: np.ndarray, weights: np.ndarray
) -> np.ndarray:
    """
    Computes Moment of inertia tensor for a batch of coordinates.
    :param coord: (B, N, 3)
    :param weights: (B, N) or (N,)
    :return: (B, 3, 3)
    """
    coord = coord.astype(np.float32, copy=False)
    weights = weights.astype(coord.dtype, copy=False)

    x = coord[..., 0]  # (B, N)
    y = coord[..., 1]
    z = coord[..., 2]

    i_xx = (weights * (y**2 + z**2)).sum(axis=1)  # (B,)
    i_yy = (weights * (x**2 + z**2)).sum(axis=1)
    i_zz = (weights * (x**2 + y**2)).sum(axis=1)

    i_xy = -(weights * (x * y)).sum(axis=1)
    i_xz = -(weights * (x * z)).sum(axis=1)
    i_yz = -(weights * (y * z)).sum(axis=1)

    batch_size = coord.shape[0]
    moi = np.zeros((batch_size, 3, 3), dtype=coord.dtype)

    moi[:, 0, 0] = i_xx
    moi[:, 1, 1] = i_yy
    moi[:, 2, 2] = i_zz

    moi[:, 0, 1] = moi[:, 1, 0] = i_xy
    moi[:, 0, 2] = moi[:, 2, 0] = i_xz
    moi[:, 1, 2] = moi[:, 2, 1] = i_yz

    return moi


def align_mol_to_principal_frame_onnx(mol):
    """
    Align a molecule to its principal inertial frame.
    :param mol: rdkit Mol with a conformer
    :return: (context, shift, rotation, aligned_coord)
    """
    conf = mol.GetConformer()
    coord = np.array(conf.GetPositions(), dtype=np.float32)

    virtual_com = np.mean(coord, axis=0)
    coord = coord - virtual_com

    shift = -virtual_com

    context, aligned_coord, rotation = get_context_shape_onnx(
        coord, include_rotation=True
    )

    return context, shift, rotation, aligned_coord


# TODO: Double Check
def concat_masked_and_pad_onnx(
    xs: list[np.ndarray],
    masks: list[np.ndarray],
    pad_extra: int = 0,
    pad_to: int | None = None,
    pad_value: float = 0.0,
):
    """
    xs:    list of arrays, each (B, N, *D)
    masks: list of masks, each (B, N, 1) or (B, N) (bool or 0/1)

    Returns: (B, L, *D), where L depends on pad_extra / pad_to.
    """

    # (B, K, N, *D)
    x = np.stack(xs, axis=1)

    # (B, K, N, 1) or (B, K, N)
    m = np.stack(masks, axis=1)
    if m.ndim == x.ndim:  # mask has trailing singleton
        m = np.squeeze(m, axis=-1)

    m = m.astype(bool)

    b, k, n = m.shape
    feat_shape = x.shape[3:]  # (*D)
    flat_len = k * n

    # Flatten K and N -> (B, K*N, *D)
    x_f = x.reshape(b, flat_len, *feat_shape)
    m_f = m.reshape(b, flat_len)

    # Select ragged per batch
    selected = [x_f[i][m_f[i]] for i in range(b)]

    # Ensure empty selections have correct shape
    empty = np.zeros((0, *feat_shape), dtype=x.dtype)
    selected = [t if t.size else empty for t in selected]

    # Determine max length
    Lmax = max(t.shape[0] for t in selected)

    # Pad to max length
    out = np.full((b, Lmax, *feat_shape), pad_value, dtype=x.dtype)

    for i, t in enumerate(selected):
        Li = t.shape[0]
        if Li > 0:
            out[i, :Li] = t

    # Decide final length
    if pad_to is not None:
        L = pad_to
    else:
        L = out.shape[1] + pad_extra

    # Pad or truncate
    if out.shape[1] < L:
        pad = np.full((b, L - out.shape[1], *feat_shape), pad_value, dtype=x.dtype)
        out = np.concatenate([out, pad], axis=1)
    else:
        out = out[:, :L]

    return out



import numpy as np
from rdkit import Chem
from rdkit.Chem import rdmolops

from ..utils.common import (allowable_features, bond_type_dict,
                            elements_decoder, elements_dict)
from ..utils.config import DIMENSION, NUM_BOND_TYPES


class MolGraphONNX:
    """
    A class to handle molecular graphs without PyTorch:
    """

    def __init__(self, x: np.ndarray, edge_index: np.ndarray, edge_attr: np.ndarray):
        self.x = x
        self.edge_index = edge_index
        self.edge_attr = edge_attr

    @classmethod
    def from_mol(cls, mol: Chem.Mol, remove_hs: bool = True) -> "MolGraphONNX":
        """
        Converts rdkit mol object to MolGraph object
        geometric package. Strips hydrogens from the Mol object. Ignores Atoms and Bonds Chirality,
        Bonds are represented in edge_attrs as integers:
        1 - Single
        2 - Double
        3 - Triple
        4 - Aromatic
        :param mol: rdkit mol object
        :param remove_hs: if H atoms are to be removed
        :return: graph data object with the attributes: x, edge_index, edge_attr
        """
        # Remove hydrogens from the molecule - to simplify graph structure. Ids of atoms remain unchanged.
        if remove_hs:
            mol = rdmolops.RemoveHs(mol)

        out = [0] * len(mol.GetAtoms())
        for atom in mol.GetAtoms():
            element = atom.GetAtomicNum()
            index = atom.GetIdx()
            out[index] = element

        x = np.array(out, dtype=np.float32)

        # bonds
        if len(mol.GetBonds()) > 0:  # mol has bonds
            edges_list = []
            edge_features_list = []
            for bond in mol.GetBonds():
                i = bond.GetBeginAtomIdx()
                j = bond.GetEndAtomIdx()
                edge_feature = (
                    allowable_features["possible_bonds"].index(bond.GetBondType()) + 1
                )
                edges_list.append((i, j))
                edge_features_list.append(edge_feature)
                edges_list.append((j, i))
                edge_features_list.append(edge_feature)

            # data.edge_index: Graph connectivity in COO format with shape [2, num_edges]
            edge_index = np.array(edges_list, dtype=np.int64).T

            # data.edge_attr: Edge feature matrix with shape [num_edges, num_edge_features]
            edge_attr = np.array(edge_features_list, dtype=np.float32)

        else:  # mol has no bonds
            raise ValueError(
                f"Bonds must be specified for the molecule - {mol.GetProp('_Name')}."
            )

        return cls(x=x, edge_index=edge_index, edge_attr=edge_attr)

    def adjacency_matrix(self, padded: bool = True) -> np.ndarray:
        """
        Creates a 0-1 normalised adjacency matrix with a specified size from a MolGraph object
        representing a molecule. Bond types are represented as follows:
        0 - No Bond
        1 - Single
        2 - Double
        3 - Triple
        4 - Aromatic
        :return: adjacency matrix of a restricted shape as np ndarray
        """
        graph_size = len(self.x)
        bonds_size = len(self.edge_attr)

        if padded:
            adjacency_matrix = np.zeros(
                (DIMENSION, DIMENSION, NUM_BOND_TYPES), dtype=np.float32
            )
        else:
            adjacency_matrix = np.zeros(
                (graph_size, graph_size, NUM_BOND_TYPES), dtype=np.float32
            )

        adjacency_matrix[:, :, 0] = 1

        if graph_size > DIMENSION:
            raise ValueError(f"The graph should have not more than {DIMENSION} nodes")
        if self.edge_attr is None:
            raise ValueError(f"Bond types should be specified in edge_attr of Data")

        edge_attr = self.edge_attr.astype(np.int64)

        for i in range(bonds_size):
            x = self.edge_index[0][i]
            y = self.edge_index[1][i]

            adjacency_matrix[x][y][0] = 0
            adjacency_matrix[y][x][0] = 0

            adjacency_matrix[x][y][edge_attr[i]] = 1
            adjacency_matrix[y][x][edge_attr[i]] = 1

        return adjacency_matrix

    def to_rdkit_mol(self):
        rw_mol = Chem.RWMol()
        atom_indexes = []

        atoms = self.x.tolist()
        bond_index = self.edge_index.tolist()
        bond_attr = self.edge_attr.tolist()

        for atom in atoms:
            idx = rw_mol.AddAtom(Chem.Atom(elements_dict[atom[0]]))
            atom_indexes.append(idx)

        for i, bond in enumerate(bond_index[0]):
            try:
                rw_mol.AddBond(
                    atom_indexes[bond_index[0][i]],
                    atom_indexes[bond_index[1][i]],
                    bond_type_dict[bond_attr[i]],
                )
            except Exception:
                pass

        mol = rw_mol.GetMol()
        return mol

    def elements_vector(self) -> np.ndarray:
        """
        Returns a fixed-sized elements vector
        :return: [atomic_num, ...0...] size(DIMENSION, 1)
        """
        elements_vector = np.zeros(DIMENSION, dtype=np.int64)

        for i in range(len(self.x)):
            elements_vector[i] = self.x[i]

        return elements_vector

    def one_hot_elements_encoding(self, max_n_nodes) -> np.ndarray:
        """
        Returns a one-hot encoded fixed-sized elements vector;
        the number of types is the length of PERMITTED ELEMENTS set
        :return: [, ...0...] size(DIMENSION, len(PERMITTED_ELEMENTS), 1)
        """
        one_hot = np.zeros((max_n_nodes, len(elements_decoder.keys())), dtype=np.int64)

        for i in range(len(self.x)):
            atom_type = elements_decoder[self.x[i].item()]
            one_hot[i][atom_type] = 1

        return one_hot



from typing import Tuple

import numpy
import torch
from rdkit import Chem
from rdkit.Chem import rdmolops

from .common import allowable_features
from .common import bond_type_dict as bonds_dict
from .common import elements_decoder, elements_dict
from .config import DIMENSION, NUM_BOND_TYPES


class MolGraph:
    """
    A class to handle molecular graphs using torch.Tensors:
    """

    def __init__(
        self, x: torch.Tensor, edge_index: torch.Tensor, edge_attr: torch.Tensor
    ):
        self.x = x
        self.edge_index = edge_index
        self.edge_attr = edge_attr

    @classmethod
    def from_adjacency_matrix(
        cls,
        nodes: torch.Tensor,
        adjacency_matrix: torch.Tensor,
    ) -> "MolGraph":
        """
        Create a Molecular Graph from Nodes tensor and given 0-1 normalised adjacency matrix of a proper dimension.
        :param nodes: [[atomic_num, shielding]..] dtype = torch.float
        :param adjacency_matrix: torch tensor of size (DIMENSION, DIMENSION)
        :return: MolGraph object
        """
        if nodes is None:
            raise ValueError(f"Either Nodes tensor or Atom Matrix should be specified.")

        n = len(nodes)

        if adjacency_matrix is None:
            raise ValueError(f"Adjacency matrix should be Specified")

        if adjacency_matrix.size() != torch.Size(
            [DIMENSION, DIMENSION, NUM_BOND_TYPES]
        ):
            raise ValueError(
                f"Adjacency matrix should be of size {DIMENSION} with bond encoding with size of {NUM_BOND_TYPES}"
            )

        repr_m = torch.argmax(adjacency_matrix, dim=2)[:n, :n]
        mask = repr_m != 0
        rows, cols = torch.where(mask)
        edge_index = torch.stack([rows, cols])
        edge_attr = repr_m[mask]

        return cls(
            x=nodes,
            edge_index=edge_index,
            edge_attr=edge_attr,
        )

    @classmethod
    def from_mol(cls, mol: Chem.Mol, remove_hs: bool = True) -> "MolGraph":
        """
        Converts rdkit mol object to MolGraph object
        geometric package. Strips hydrogens from the Mol object. Ignores Atoms and Bonds Chirality,
        Bonds are represented in edge_attrs as integers:
        1 - Single
        2 - Double
        3 - Triple
        4 - Aromatic
        :param mol: rdkit mol object
        :param remove_hs: if H atoms are to be removed
        :return: graph data object with the attributes: x, edge_index, edge_attr
        """
        # Remove hydrogens from the molecule - to simplify graph structure. Ids of atoms remain unchanged.
        if remove_hs:
            mol = rdmolops.RemoveHs(mol)

        out = [0] * len(mol.GetAtoms())
        for atom in mol.GetAtoms():
            element = atom.GetAtomicNum()
            index = atom.GetIdx()
            out[index] = element

        x = torch.tensor(out, dtype=torch.float)

        # bonds
        if len(mol.GetBonds()) > 0:  # mol has bonds
            edges_list = []
            edge_features_list = []
            for bond in mol.GetBonds():
                i = bond.GetBeginAtomIdx()
                j = bond.GetEndAtomIdx()
                edge_feature = (
                    allowable_features["possible_bonds"].index(bond.GetBondType()) + 1
                )
                edges_list.append((i, j))
                edge_features_list.append(edge_feature)
                edges_list.append((j, i))
                edge_features_list.append(edge_feature)

            # data.edge_index: Graph connectivity in COO format with shape [2, num_edges]
            edge_index = torch.tensor(numpy.array(edges_list).T, dtype=torch.long)

            # data.edge_attr: Edge feature matrix with shape [num_edges, num_edge_features]
            edge_attr = torch.tensor(
                numpy.array(edge_features_list, dtype=float), dtype=torch.float
            )
        else:  # mol has no bonds
            raise ValueError(
                f"Bonds must be specified for the molecule - {mol.GetProp('_Name')}."
            )

        return cls(x=x, edge_index=edge_index, edge_attr=edge_attr)

    def adjacency_matrix(self, padded: bool = True) -> torch.Tensor:
        """
        Creates a 0-1 normalised adjacency matrix with a specified size from a MolGraph object
        representing a molecule. Bond types are represented as follows:
        0 - No Bond
        1 - Single
        2 - Double
        3 - Triple
        4 - Aromatic
        :return: adjacency matrix of a restricted shape as torch tensor
        """
        graph_size = len(self.x)
        bonds_size = len(self.edge_attr)

        device = self.x.device
        if padded:
            adjacency_matrix = torch.zeros(
                DIMENSION, DIMENSION, NUM_BOND_TYPES, dtype=torch.float, device=device
            )
        else:
            adjacency_matrix = torch.zeros(
                graph_size, graph_size, NUM_BOND_TYPES, dtype=torch.float, device=device
            )

        adjacency_matrix[:, :, 0] = 1

        if graph_size > DIMENSION:
            raise ValueError(f"The graph should have not more than {DIMENSION} nodes")
        if self.edge_attr is None:
            raise ValueError(f"Bond types should be specified in edge_attr of Data")

        for i in range(bonds_size):
            x = self.edge_index[0][i]
            y = self.edge_index[1][i]

            adjacency_matrix[x][y][0] = 0
            adjacency_matrix[y][x][0] = 0

            adjacency_matrix[x][y][self.edge_attr[i].long().item()] = 1
            adjacency_matrix[y][x][self.edge_attr[i].long().item()] = 1

        return adjacency_matrix

    def to_rdkit_mol(self):
        rw_mol = Chem.RWMol()
        atom_indexes = []

        atoms = self.x.tolist()
        bond_index = self.edge_index.tolist()
        bond_attr = self.edge_attr.tolist()

        for atom in atoms:
            idx = rw_mol.AddAtom(Chem.Atom(elements_dict[atom[0]]))
            atom_indexes.append(idx)

        for i, bond in enumerate(bond_index[0]):
            try:
                rw_mol.AddBond(
                    atom_indexes[bond_index[0][i]],
                    atom_indexes[bond_index[1][i]],
                    bonds_dict[bond_attr[i]],
                )
            except Exception:
                pass

        mol = rw_mol.GetMol()
        return mol

    def elements_vector(self) -> torch.Tensor:
        """
        Returns a fixed-sized elements vector
        :return: [atomic_num, ...0...] size(DIMENSION, 1)
        """
        elements_vector = torch.zeros(DIMENSION, dtype=torch.long, device=self.x.device)
        elements_vector[: len(self.x)] = self.x.to(torch.long)
        return elements_vector

    def one_hot_elements_encoding(self, max_n_nodes) -> torch.Tensor:
        """
        Returns a one-hot encoded fixed-sized elements vector;
        the number of types is the length of PERMITTED ELEMENTS set
        :return: [, ...0...] size(DIMENSION, len(PERMITTED_ELEMENTS), 1)
        """
        one_hot = torch.zeros(
            max_n_nodes, len(elements_decoder), dtype=torch.long, device=self.x.device
        )

        atom_types = torch.tensor(
            [elements_decoder[int(a)] for a in self.x],
            dtype=torch.long,
            device=self.x.device,
        )

        one_hot[torch.arange(len(self.x), device=self.x.device), atom_types] = 1

        return one_hot


def vector_graph_sort(
    elements: torch.Tensor,
    coordinates: torch.Tensor,
    adjacency_matrix: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    A function to efficiently sort the graph nodes order and
    rebuilds the adjacency matrix according to the sorted order
    :param elements: torch tensor (BATCH_SIZE, DIMENSION) -> elements vector
    :param coordinates: torch tensor (BATCH_SIZE, DIMENSION, 3) -> coordinates
    :param adjacency_matrix: (BATCH_SIZE, DIMENSION, DIMENSION, NUM_BOND_TYPES) ->
                                      adjacency matrix corresponding to order
    :return: sorted atoms_matrix and adjacency matrix as a tuple
    """
    # Sort the molecular graphs after generating shielding on GPU
    size = adjacency_matrix.size()
    batch_size = size[0]
    N = size[1]
    num_classes = size[-1]

    # Calculate distances to each atom from the center of mass (oroging)
    distances = torch.sum(torch.pow(coordinates, 2), dim=-1)

    nodes_batch = torch.stack((elements, distances), 2)
    sorted_indexes = torch.sort(torch.sum(nodes_batch, dim=2), descending=True)[1]

    sorted_elements = elements[torch.arange(batch_size).unsqueeze(1), sorted_indexes]
    sorted_coordinates = coordinates[
        torch.arange(batch_size).unsqueeze(1), sorted_indexes
    ]

    # Rearrange the adjacency matrix
    flattened_adjacency_matrix = torch.argmax(adjacency_matrix, dim=3)
    findex = sorted_indexes.repeat_interleave(N, 1) * N + sorted_indexes.repeat(1, N)
    sorted_matrix = (
        flattened_adjacency_matrix.flatten(1, 2)
        .gather(1, findex)
        .view(batch_size, N, N)
    )

    target = torch.nn.functional.one_hot(sorted_matrix, num_classes).type(torch.float)

    return sorted_elements, sorted_coordinates, target



from .cheminformatics import evaluate_samples
from .conformer_generator import MLConformerGenerator
from .inertial_fragment_matching import (ff_inertial_fragment_matching,
                                         inertial_fragment_matching)
from .onnx import MLConformerGeneratorONNX

__all__ = [
    "MLConformerGenerator",
    "MLConformerGeneratorONNX",
    "evaluate_samples",
    "inertial_fragment_matching",
    "ff_inertial_fragment_matching",
]



from .common import (apply_transform, bond_type_dict, canonicalise,
                     elements_decoder, set_conformer_positions)
from .config import (ATOM_DECODER, CONTEXT_NORMS, DIMENSION, MAX_FRAG_SIZE,
                     MAX_N_NODES, MIN_FRAG_SIZE, MIN_N_NODES, NUM_BOND_TYPES)
from .mol_split import extract_fragment, split_molecule_size_constrained
from .mol_utils import (align_mol_to_principal_frame, concat_masked_and_pad,
                        coord_to_pf_batched, get_context_shape,
                        get_moment_of_inertia_tensor, ifm_get_xh_from_fragment,
                        ifm_prepare_fragments_for_merge,
                        ifm_prepare_gen_fragment_context,
                        inverse_coord_transform, prepare_adj_mat_seer_input,
                        prepare_edm_input, prepare_fragment, prepare_masks,
                        redefine_bonds, samples_to_rdkit_mol)
from .standardizer import standardize_mol

__all__ = [
    "align_mol_to_principal_frame",
    "extract_fragment",
    "samples_to_rdkit_mol",
    "set_conformer_positions",
    "split_molecule_size_constrained",
    "apply_transform",
    "bond_type_dict",
    "canonicalise",
    "elements_decoder",
    "coord_to_pf_batched",
    "concat_masked_and_pad",
    "get_context_shape",
    "get_moment_of_inertia_tensor",
    "prepare_adj_mat_seer_input",
    "prepare_edm_input",
    "prepare_fragment",
    "prepare_masks",
    "ifm_prepare_gen_fragment_context",
    "ifm_prepare_fragments_for_merge",
    "ifm_get_xh_from_fragment",
    "inverse_coord_transform",
    "redefine_bonds",
    "standardize_mol",
    "DIMENSION",
    "NUM_BOND_TYPES",
    "MIN_N_NODES",
    "MAX_N_NODES",
    "CONTEXT_NORMS",
    "ATOM_DECODER",
    "MIN_FRAG_SIZE",
    "MAX_FRAG_SIZE",
]



from typing import Tuple, Union

import numpy as np


def clip_noise_schedule(alphas2: np.ndarray, clip_value: float = 0.001) -> np.ndarray:
    """
    For a noise schedule given by alpha^2, this clips alpha_t / alpha_t-1. This may help improve stability during
    sampling.
    """

    alphas2 = np.concatenate((np.ones(1), alphas2), axis=0)

    alphas_step = alphas2[1:] / alphas2[:-1]

    alphas_step = np.clip(alphas_step, a_min=clip_value, a_max=1.0)
    alphas2 = np.cumprod(alphas_step, axis=0)

    return alphas2


def polynomial_schedule(timesteps: int, s: float = 1e-4, power: int = 2):
    """
    A noise schedule based on a simple polynomial equation: 1 - x^power.

    Remark - rewritten in torch only
    """
    steps = timesteps + 1
    x = np.linspace(0, steps, steps)
    alphas2 = (1 - np.power(x / steps, power)) ** 2

    alphas2 = clip_noise_schedule(alphas2, clip_value=0.001)

    precision = 1 - 2 * s

    alphas2 = precision * alphas2 + s

    return alphas2


def remove_mean_with_mask(x, node_mask):
    n = np.sum(node_mask, 1, keepdims=True)

    mean = np.sum(x, 1, keepdims=True) / n
    x = x - mean * node_mask
    return x


def sample_center_gravity_zero_gaussian_with_mask(
    size: Tuple[int, int, int], node_mask: np.ndarray
):
    x = np.random.randn(size[0], size[1], size[2])

    x_masked = x * node_mask

    # This projection only works because Gaussian is rotation invariant around
    # zero and samples are independent
    x_projected = remove_mean_with_mask(x_masked, node_mask)
    return x_projected


def sample_gaussian_with_mask(size: Tuple[int, int, int], node_mask: np.ndarray):
    x = np.random.randn(size[0], size[1], size[2])

    x_masked = x * node_mask
    return x_masked


def align_fragment_com_to_generated(
    z_known_noised: np.ndarray, z_generated: np.ndarray, fixed_mask: np.ndarray
) -> np.ndarray:
    """
    Aligns COM of the fixed fragment with the corresponding generated fragment during inpainting for equivariance.
    :param z_known_noised: z_known with noise applied
    :param z_generated: z_generated with comparable nois
    :param fixed_mask: a mask to indentify the fixed fragment
    :return: aligned latent representation of a fixed fragment
    """

    coords_known = z_known_noised[:, :, :3]
    coords_gen = z_generated[:, :, :3]

    frag_com_gen = np.sum(coords_gen * fixed_mask, axis=1, keepdims=True) / (
        np.sum(fixed_mask, axis=1, keepdims=True)
    )
    frag_com_known = np.sum(coords_known * fixed_mask, axis=1, keepdims=True) / (
        np.sum(fixed_mask, axis=1, keepdims=True)
    )

    shift = frag_com_gen - frag_com_known
    coords_shifted = coords_known + shift * fixed_mask  # only move fixed region

    z_known_shifted = z_known_noised.copy()
    z_known_shifted[:, :, :3] = coords_shifted
    return z_known_shifted


class PredefinedNoiseSchedule:
    """
    Predefined noise schedule. Essentially creates a lookup array for predefined (non-learned) noise schedules.
    """

    def __init__(self, timesteps: int, precision: float, power: int = 2):
        self.timesteps = timesteps

        # Default Schedule - polynomial with power 2

        alphas2 = polynomial_schedule(timesteps, s=precision, power=power)

        sigmas2 = 1 - alphas2

        log_alphas2 = np.log(alphas2)
        log_sigmas2 = np.log(sigmas2)

        log_alphas2_to_sigmas2 = log_alphas2 - log_sigmas2

        self.gamma = (-log_alphas2_to_sigmas2).astype(np.float32)

    def __call__(self, t: np.ndarray):
        t_int = np.round(t * self.timesteps).astype(int)
        return self.gamma[t_int]


class EquivariantDiffusionONNX:
    """
    PyTorch - free implementation of
    The E(n) Diffusion Module.
    """

    def __init__(
        self,
        egnn_onnx: str,
        in_node_nf: int = 8,
        n_dims: int = 3,
        timesteps: int = 1000,
        noise_precision: float = 1e-4,
        norm_values: Tuple[float, float] = (
            1.0,
            9.0,
        ),  # (1, max number of atom classes)
    ):
        try:
            import onnxruntime
        except ImportError as e:
            raise ImportError(
                'Failed to import onnxruntime. To resolve run `pip install "mlconfgen[onnx]"`\n'
            ) from e

        super().__init__()

        self.gamma = PredefinedNoiseSchedule(
            timesteps=timesteps, precision=noise_precision
        )

        # The network that will predict the denoising.

        self.dynamics = onnxruntime.InferenceSession(egnn_onnx)

        self.in_node_nf = in_node_nf
        self.n_dims = n_dims
        self.num_classes = self.in_node_nf

        # Declare time steps-related tensors
        self.T = timesteps
        self.time_steps = np.flip(np.arange(0, timesteps))

        self.norm_values = norm_values

    def phi(self, x, t, node_mask, edge_mask, context):
        inputs = {
            "t": t.astype(np.float32),
            "xh": x.astype(np.float32),
            "node_mask": node_mask.astype(np.float32),
            "edge_mask": edge_mask.astype(np.float32),
            "context": context.astype(np.float32),
        }
        net_out = self.dynamics.run(None, inputs)

        return net_out[0]

    @staticmethod
    def sigmoid(z: Union[np.ndarray, float]):
        return 1 / (1 + np.exp(-z))

    @staticmethod
    def softplus(z: Union[np.ndarray, float]):
        return np.log1p(np.exp(z))

    @staticmethod
    def logsigmoid(z: Union[np.ndarray, float]):
        return -np.log1p(np.exp(-z))

    @staticmethod
    def one_hot(labels: np.ndarray, num_classes: int):
        return np.eye(num_classes)[labels]

    @staticmethod
    def inflate_batch_array(array: np.ndarray, target: np.ndarray) -> np.ndarray:
        """
        Inflates the batch array (array) with only a single axis (i.e. shape = (batch_size,), or possibly more empty
        axes (i.e. shape (batch_size, 1, ..., 1)) to match the target shape.
        """
        target_shape = (array.shape[0],) + (1,) * (len(target.shape) - 1)
        return array.reshape(target_shape)

    def sigma(self, gamma: np.ndarray, target_tensor: np.ndarray) -> np.ndarray:
        """Computes sigma given gamma."""
        return self.inflate_batch_array(np.sqrt(self.sigmoid(gamma)), target_tensor)

    def alpha(self, gamma: np.ndarray, target_tensor: np.ndarray):
        """Computes alpha given gamma."""
        return self.inflate_batch_array(np.sqrt(self.sigmoid(-gamma)), target_tensor)

    @staticmethod
    def snr(gamma: np.ndarray) -> np.ndarray:
        """Computes signal to noise ratio (alpha^2/sigma^2) given gamma."""
        return np.exp(-gamma)

    def unnormalize(
        self, x: np.ndarray, h_cat: np.ndarray, node_mask: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        x = x * self.norm_values[0]
        h_cat = h_cat * self.norm_values[1]

        h_cat = h_cat * node_mask

        return x, h_cat

    def sigma_and_alpha_t_given_s(
        self, gamma_t: np.ndarray, gamma_s: np.ndarray, target_tensor: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes sigma t given s, using gamma_t and gamma_s. Used during sampling.

        These are defined as:
            alpha t given s = alpha t / alpha s,
            sigma t given s = sqrt(1 - (alpha t given s) ^2 ).
        """
        sigma2_t_given_s = self.inflate_batch_array(
            1 - np.exp(self.softplus(gamma_s) - self.softplus(gamma_t)), target_tensor
        )

        log_alpha2_t = self.logsigmoid(-gamma_t)
        log_alpha2_s = self.logsigmoid(-gamma_s)
        log_alpha2_t_given_s = log_alpha2_t - log_alpha2_s

        alpha_t_given_s = np.exp(0.5 * log_alpha2_t_given_s)
        alpha_t_given_s = self.inflate_batch_array(alpha_t_given_s, target_tensor)

        sigma_t_given_s = np.sqrt(sigma2_t_given_s)

        return sigma2_t_given_s, sigma_t_given_s, alpha_t_given_s

    def compute_x_pred(
        self, net_out: np.ndarray, zt: np.ndarray, gamma_t: np.ndarray
    ) -> np.ndarray:
        """Commputes x_pred, i.e. the most likely prediction of x."""

        sigma_t = self.sigma(gamma_t, target_tensor=net_out)
        alpha_t = self.alpha(gamma_t, target_tensor=net_out)
        eps_t = net_out
        x_pred = 1.0 / alpha_t * (zt - sigma_t * eps_t)

        return x_pred

    def sample_p_xh_given_z0(
        self,
        z0: np.ndarray,
        node_mask: np.ndarray,
        edge_mask: np.ndarray,
        context: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Samples x ~ p(x|z0)."""
        zeros = np.zeros(shape=(z0.shape[0], 1))
        gamma_0 = self.gamma(zeros)
        # Computes sqrt(sigma_0^2 / alpha_0^2)
        sigma_x = np.expand_dims(self.snr(-0.5 * gamma_0), 1)
        net_out = self.phi(z0, zeros, node_mask, edge_mask, context)

        # Compute mu for p(zs | zt).
        mu_x = self.compute_x_pred(net_out, z0, gamma_0)
        xh = self.sample_normal(mu=mu_x, sigma=sigma_x, node_mask=node_mask)

        x = xh[:, :, : self.n_dims]

        x, h_cat = self.unnormalize(x, z0[:, :, self.n_dims : -1], node_mask)

        h_cat = self.one_hot(np.argmax(h_cat, axis=2), self.num_classes) * node_mask
        h = h_cat
        return x, h

    def sample_normal(
        self, mu: np.ndarray, sigma: np.ndarray, node_mask: np.ndarray
    ) -> np.ndarray:
        """Samples from a Normal distribution."""
        bs = mu.shape[0]
        eps = self.sample_combined_position_feature_noise(bs, mu.shape[1], node_mask)
        return mu + sigma * eps

    def sample_p_zs_given_zt(
        self,
        s: np.ndarray,
        t: np.ndarray,
        zt: np.ndarray,
        node_mask: np.ndarray,
        edge_mask: np.ndarray,
        context: np.ndarray,
    ) -> np.ndarray:
        """Samples from zs ~ p(zs | zt). Only used during sampling."""
        gamma_s = self.gamma(s)
        gamma_t = self.gamma(t)

        (
            sigma2_t_given_s,
            sigma_t_given_s,
            alpha_t_given_s,
        ) = self.sigma_and_alpha_t_given_s(gamma_t, gamma_s, zt)

        sigma_s = self.sigma(gamma_s, target_tensor=zt)
        sigma_t = self.sigma(gamma_t, target_tensor=zt)

        # Neural net prediction.
        eps_t = self.phi(zt, t, node_mask, edge_mask, context)

        mu = (
            zt / alpha_t_given_s
            - (sigma2_t_given_s / alpha_t_given_s / sigma_t) * eps_t
        )

        # Compute sigma for p(zs | zt).
        sigma = sigma_t_given_s * sigma_s / sigma_t

        # Sample zs given the parameters derived from zt.
        zs = self.sample_normal(mu, sigma, node_mask)

        # Project down to avoid numerical runaway of the center of gravity.
        zs = np.concatenate(
            [
                remove_mean_with_mask(zs[:, :, : self.n_dims], node_mask),
                zs[:, :, self.n_dims :],
            ],
            axis=2,
        )
        return zs

    def sample_combined_position_feature_noise(
        self, n_samples: int, n_nodes: int, node_mask: np.ndarray
    ) -> np.ndarray:
        """
        Samples mean-centered normal noise for z_x, and standard normal noise for z_h.
        """
        z_x = sample_center_gravity_zero_gaussian_with_mask(
            size=(n_samples, n_nodes, self.n_dims),
            node_mask=node_mask,
        )

        z_h = sample_gaussian_with_mask(
            size=(
                n_samples,
                n_nodes,
                self.in_node_nf,
            ),
            node_mask=node_mask,
        )

        z = np.concatenate([z_x, z_h], axis=2)
        return z

    def __call__(
        self,
        node_mask: np.ndarray,
        edge_mask: np.ndarray,
        context: np.ndarray,
        resample_steps: int = 0,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Draw samples from the generative model.
        Inference

        :param node_mask: node mask tensor
        :param edge_mask: edge mask tensor
        :param context: batched context for generation
        :param resample_steps: number of resampling steps for harmonisation
        :return: generated samples in tensor representation
        """
        n_samples, n_nodes, _ = node_mask.shape

        z = self.sample_combined_position_feature_noise(n_samples, n_nodes, node_mask)

        # Iteratively sample p(z_s | z_t) for t = 1, ..., T, with s = t - 1.
        for s in self.time_steps:
            s_array = np.full([n_samples, 1], fill_value=s)
            t_array = s_array + 1.0
            s_array = s_array / self.T
            t_array = t_array / self.T

            # Optional Resampling loop for improvement of generation quality
            for _ in range(resample_steps):
                z = self.sample_p_zs_given_zt(
                    s_array,
                    t_array,
                    z,
                    node_mask,
                    edge_mask,
                    context,
                )

            z = self.sample_p_zs_given_zt(
                s_array,
                t_array,
                z,
                node_mask,
                edge_mask,
                context,
            )

        # Finally sample p(x, h | z_0).
        x, h = self.sample_p_xh_given_z0(
            z,
            node_mask,
            edge_mask,
            context,
        )

        return x, h

    def inpaint(
        self,
        node_mask: np.ndarray,
        edge_mask: np.ndarray,
        context: np.ndarray,
        z_known: np.ndarray,
        fixed_mask: np.ndarray,
        resample_steps: int = 10,
        blend_power: int = 3,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Draw samples from the generative model while fixing a given fragment
        Inference

        :param node_mask: node mask tensor
        :param edge_mask: edge mask tensor
        :param context: batched context for generation
        :param z_known: latent representation of a fixed fragment
        :param fixed_mask: mask to indicate the position of fixed atoms
        :param resample_steps: number of resampling steps for harmonisation
        :param blend_power: power of the polynomial blending schedule
        :return: generated samples in tensor representation
        """
        n_samples, n_nodes, _ = node_mask.shape

        z = self.sample_combined_position_feature_noise(n_samples, n_nodes, node_mask)

        # Iteratively sample p(z_s | z_t) for t = 1, ..., T, with s = t - 1.
        for s in self.time_steps:
            s_array = np.full([n_samples, 1], fill_value=s)
            t_array = s_array + 1.0
            s_array = s_array / self.T
            t_array = t_array / self.T

            # Polynomial blending
            blend = np.power((1 - s_array), blend_power).reshape(n_samples, 1, 1)

            for _ in range(resample_steps):
                z = self.sample_p_zs_given_zt(
                    s_array,
                    t_array,
                    z,
                    node_mask,
                    edge_mask,
                    context,
                )

                # Forward-diffuse the known fragment at timestep s
                gamma_s = self.gamma(s_array)
                alpha_s = self.alpha(gamma_s, z_known)
                sigma_s = self.sigma(gamma_s, z_known)

                eps_frag = self.sample_combined_position_feature_noise(
                    n_samples, n_nodes, node_mask
                )
                z_known_noised = alpha_s * z_known + sigma_s * eps_frag

                # COM alignment
                z_known_noised = align_fragment_com_to_generated(
                    z_known_noised, z, fixed_mask
                )

                # Softly blend fragment into z
                z = (
                    blend * z_known_noised * fixed_mask
                    + (1 - blend) * z * fixed_mask
                    + z * (1 - fixed_mask)
                )

            z = self.sample_p_zs_given_zt(
                s_array,
                t_array,
                z,
                node_mask,
                edge_mask,
                context,
            )

        # Finally sample p(x, h | z_0).
        x, h = self.sample_p_xh_given_z0(
            z,
            node_mask,
            edge_mask,
            context,
        )

        return x, h

    def ifm_merge_fragments(
        self,
        node_mask: np.ndarray,
        edge_mask: np.ndarray,
        context: np.ndarray,
        z_seed: np.ndarray,
        diffusion_level: int = 10,
        resample_steps: int = 1,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Merges fragments, while allowing all of them to be adjusted by the model.
        Inference

        :param node_mask: node mask tensor
        :param edge_mask: edge mask tensor
        :param context: batched context for generation
        :param z_seed: latent representation of unmerged fragments
        :param diffusion_level: a depth of diffusion to be applied during merging
        :param resample_steps: number of resampling steps for harmonisation
        :return: generated samples in tensor representation
        """
        if resample_steps < 1:
            resample_steps = 1

        n_samples, n_nodes, _ = node_mask.size()

        # Works better without forward Diffusion!
        z = z_seed

        for s in self.time_steps:
            if s > diffusion_level:
                continue

            s_array = np.full([n_samples, 1], fill_value=s)
            t_array = s_array + 1.0
            s_array = s_array / self.T
            t_array = t_array / self.T

            # Polynomial blending

            for _ in range(resample_steps):
                z = self.sample_p_zs_given_zt(
                    s_array,
                    t_array,
                    z,
                    node_mask,
                    edge_mask,
                    context,
                )

            z = self.sample_p_zs_given_zt(
                s_array,
                t_array,
                z,
                node_mask,
                edge_mask,
                context,
            )

        # Decode
        x, h = self.sample_p_xh_given_z0(z, node_mask, edge_mask, context)
        return x, h

    def ifm_merge_fragments_with_injection(
        self,
        node_mask: np.ndarray,
        edge_mask: np.ndarray,
        fixed_mask: np.ndarray,  # (B, N, 1)
        context: np.ndarray,
        z_seed: np.ndarray,
        diffusion_level: int = 10,
        resample_steps: int = 1,
        blend_power: int = 3,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Merges 2 fragments, while fixing the target one and allowing the other to be adjusted by the model.
        The fixed fragment is indicated with the fixed mask.
        Inference

        :param node_mask: node mask tensor
        :param edge_mask: edge mask tensor
        :param context: batched context for generation
        :param z_seed: latent representation of unmerged fragments
        :param fixed_mask: mask to indicate the position of fixed atoms
        :param diffusion_level: a depth of diffusion to be applied during merging
        :param resample_steps: number of resampling steps for harmonisation
        :param blend_power: power of the polynomial blending schedule
        :return: generated samples in tensor representation


        """
        if resample_steps < 1:
            resample_steps = 1

        n_samples, n_nodes, _ = node_mask.size()

        # Forward diffuse the full structure, as investigated, works a bit worse than direct z_known seed.
        s_array_0 = np.full([n_samples, 1], fill_value=diffusion_level)
        s_array_0 = s_array_0 / self.T
        gamma_s = self.gamma(s_array_0)
        alpha_s = self.alpha(gamma_s, z_seed)
        sigma_s = self.sigma(gamma_s, z_seed)

        eps = self.sample_combined_position_feature_noise(n_samples, n_nodes, node_mask)
        z_noised = alpha_s * z_seed + sigma_s * eps
        z = z_noised

        for s in self.time_steps:
            if s > diffusion_level:
                continue

            s_array = np.full([n_samples, 1], fill_value=s)
            t_array = s_array + 1.0
            s_array = s_array / self.T
            t_array = t_array / self.T

            # Polynomial blending
            blend = np.power((1 - s_array), blend_power).reshape(n_samples, 1, 1)

            for _ in range(resample_steps):
                z = self.sample_p_zs_given_zt(
                    s_array,
                    t_array,
                    z,
                    node_mask,
                    edge_mask,
                    context,
                )

                # Forward-diffuse the known fragment at timestep s
                gamma_s = self.gamma(s_array)
                alpha_s = self.alpha(gamma_s, z_seed)
                sigma_s = self.sigma(gamma_s, z_seed)

                eps_frag = self.sample_combined_position_feature_noise(
                    n_samples, n_nodes, node_mask
                )
                z_fixed_noised = alpha_s * z_seed + sigma_s * eps_frag

                # Align fixed fragment to avoid CoM drift
                z_fixed_noised = align_fragment_com_to_generated(
                    z_fixed_noised, z, fixed_mask
                )

                # Blend fixed fragment back in softly
                z = (
                    blend * z_fixed_noised * fixed_mask
                    + (1 - blend) * z * fixed_mask
                    + z * (1 - fixed_mask)
                )

        # Decode
        x, h = self.sample_p_xh_given_z0(z, node_mask, edge_mask, context)
        return x, h

    def animate(
        self,
        node_mask: np.ndarray,
        edge_mask: np.ndarray,
        context: np.ndarray,
        resample_steps: int = 0,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Iterator
        Draw samples from the generative model, while yielding all intermediate states.
        Inference

        :param node_mask: node mask tensor
        :param edge_mask: edge mask tensor
        :param context: batched context for generation
        :param resample_steps: number of resampling steps for harmonisation
        :return: generated samples in tensor representation at each denoising step
        """
        n_samples, n_nodes, _ = node_mask.size()

        z = self.sample_combined_position_feature_noise(n_samples, n_nodes, node_mask)

        # Iteratively sample p(z_s | z_t) for t = 1, ..., T, with s = t - 1.
        for s in self.time_steps:
            s_array = np.full([n_samples, 1], fill_value=s)
            t_array = s_array + 1.0
            s_array = s_array / self.T
            t_array = t_array / self.T

            # Optional Resampling loop for improvement of generation quality
            for _ in range(resample_steps):
                z = self.sample_p_zs_given_zt(
                    s_array,
                    t_array,
                    z,
                    node_mask,
                    edge_mask,
                    context,
                )

            z = self.sample_p_zs_given_zt(
                s_array,
                t_array,
                z,
                node_mask,
                edge_mask,
                context,
            )

            # Finally sample p(x, h | z_0).
            x, h = self.sample_p_xh_given_z0(
                z,
                node_mask,
                edge_mask,
                context,
            )

            yield x, h



import rdkit.Chem
import torch
from rdkit import Chem
from rdkit.Chem import rdFingerprintGenerator
from rdkit.DataStructs.cDataStructs import TanimotoSimilarity

from ..utils import set_conformer_positions
from .shape_similarity import (best_pi_rotation_by_tanimoto,
                               get_shape_quadrupole_for_molecule)

FP_SIZE = 2048
GENERATOR = rdFingerprintGenerator.GetMorganGenerator(
    radius=2, fpSize=FP_SIZE, includeChirality=False, useBondTypes=True
)


def evaluate_samples(
    reference: rdkit.Chem.Mol,
    samples: list[rdkit.Chem.Mol],
    generator: rdFingerprintGenerator = GENERATOR,
    sanitize_ref: bool = True,
) -> tuple[str, list[dict]]:
    """
    Calculate chemical and shape similarity of the generated samples to reference, while ignoring Hs
    :param reference: reference mol
    :param samples: a list of generated mols
    :param generator: fingerprint generator
    :param sanitize_ref: If reference molecule should be sanitized
    :return: molblock of a reference in a principal frame, a list of sample conformers molblocks, aligned with reference,
             along with chemical and shape tanimoto scores.
    """

    # Ensure Hs are stripped off Reference
    reference = Chem.RemoveHs(reference, sanitize=sanitize_ref)

    if sanitize_ref:
        fp_ref = generator.GetFingerprint(reference)
    else:
        fp_ref = None

    conf = reference.GetConformer()
    ref_coord = torch.tensor(conf.GetPositions(), dtype=torch.float32)

    # move coord to center
    virtual_com = torch.mean(ref_coord, dim=0)
    ref_coord = ref_coord - virtual_com

    r_s_mom, sq_ref_coord = get_shape_quadrupole_for_molecule(coordinates=ref_coord)
    # Set mol object coordinates to the principal frame
    pf_reference = set_conformer_positions(reference, sq_ref_coord)
    ref_mol_block = Chem.MolToMolBlock(pf_reference)

    results = []
    for sample in samples:
        # Calculate chemical similarity Tanimoto score
        # Ensure Hydrogens are stripped off

        # Ensure Hs are stripped off Sample
        sample = Chem.RemoveHs(sample)

        fp_sample = generator.GetFingerprint(sample)

        if sanitize_ref:
            chemical_tanimoto = TanimotoSimilarity(fp_ref, fp_sample)
        else:
            chemical_tanimoto = 0

        sample_conf = sample.GetConformer()
        sample_coord = torch.tensor(sample_conf.GetPositions(), dtype=torch.float32)

        # Move Center to COM
        s_virtual_com = torch.mean(sample_coord, dim=0)
        sample_coord = sample_coord - s_virtual_com
        s_s_mom, sq_sample_coord = get_shape_quadrupole_for_molecule(
            coordinates=sample_coord
        )

        best_coord, shape_tanimoto = best_pi_rotation_by_tanimoto(
            sq_ref_coord, sq_sample_coord
        )

        aligned_sample = set_conformer_positions(sample, best_coord)

        results.append(
            {
                "mol_block": Chem.MolToMolBlock(aligned_sample),
                "shape_tanimoto": shape_tanimoto,
                "chemical_tanimoto": chemical_tanimoto,
            }
        )
    return ref_mol_block, results
